/*++++++++++++++
.IDENTIFICATION fcat.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix
.KEYWORDS       Convert Catalogue Number
.VERSION  1.0   05-Feb-2000: List a file  whatever its compression ...
.VERSION  1.1   17-Mar-2000: Allow full directories
.VERSION  1.2   12-Mar-2001: -m write the message in a more visible way
.VERSION  1.3   30-Apr-2001: The suffix .z .Z .gz is removed.
.VERSION  1.4   26-Nov-2001: Take care of links ...
.VERSION  1.5   25-Apr-2003: Accept bzip2  (extension .bz)
.VERSION  1.6   03-Jul-2004: Take care of closed stdout
.VERSION  1.7   29-Jul-2005: Option -1: directories listed with trailing /
.VERSION  1.8   19-Jan-2006: No message in case of SIGPIPE (unless verbose mode)
.VERSION  1.9   02-Jun-2007: Cosmetics (-Wall)
.VERSION  2.0   10-Nov-2007: Signal changes
.VERSION  2.1   17-Apr-2010: Use WNOHANG in wait
.VERSION  2.2   16-Jun-2010: Accept with compressed ext if not there
.VERSION  2.3   27-Apr-2012: Option -q (quiet)
.VERSION  2.4   09-Jul-2013: Compressed FITS (.fz)
.VERSION  2.41  25-Aug-2013: Accept fimes witjout any extension
.COMMENTS       Accepts .Z .z .gz .nnn
---------------*/

#define VERSION "2.41 (2013-08-25)"

/* Problems with Fork / vfork.
   Better to use vfork which is faster, but doesn't exist on some machines...
*/
#ifdef NOvfork
#define vfork fork
#endif

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>
#include <dirent.h>
#include <fcntl.h>	/* open, close */
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/types.h>

#ifndef LEN_CWD		/* Max length of file names */
#define LEN_CWD	512
#endif

#ifndef DEBUG
#define DEBUG	0
#endif

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])
#define starts_by(s1,s2) (!strncmp(s1,s2,strlen(s2)))

static char opt1;		/* -1 opt  */
static char ohead;		/* -h opt  */
static char optf;		/* -f opt  */
static char optq;		/* -q opt  */
static char verbop   = 0; 	/* Verbose */
static int  count;
static int  skip;
static int  bs=0;		/* Blocksize (-b argument) */
static int  inSig = 0;
static char usage[] = "\
Usage: fcat [-vf1hq] [-b blocksize] [-i ignore_count] [-s max] [file...]\n\
" ;
static char help[] = "\
   -f: Just test if the file exists and can be read\n\
   -1: List the actual names of files (with/without the compression suffix)\n\
   -h: Add a message before starting each file\n\
   -b: define a blocksize for reading/writing (e.g. FITS files; default=512)\n\
   -i: number of (blocks or bytes, depending on -b option) to skip\n\
   -q: quiet option (just returns if file does not exist)\n\
   -s: stop after max (blocks or bytes, depending on -b option) written out\n\
   -v: verbose option\n\
The file can be eventually compressed, via one of the known extensions:\n\
    '.gz' (gzipped), '.bz2' (bizip2), '.fz' (FITS compress) or '.Z' (compress)\n\
" ;

static char *zcat[] = {	/* Command + Options to decompress a file */
	"gzip", "-cd", (char *)0, (char *)0
};
#define zcatfile	(ITEMS(zcat)-2)

static char *zext[] = { 	/* Extensions & Programs */
	".Z",   /* Standard Unix compress */
	".z",   /* GNU zip */
	".gz",
	".bz",  /* bzip2   */
	".bz2", /* bzip2   */
	".fz",  /* FITS rice -- Added V2.4 */
};

static int root;	/* pid of main program */
static int child ;	/* pid of child, set by open_pipe */

/* The list of files is described by a linked list */
typedef struct lname { struct lname *next; char name[2]; } LNAME ;
static LNAME *lnames ;

/*==================================================================
		Pipe utilities
 *==================================================================*/

int open_pipe(argv, mode)
/*++++++++++++++++
.PURPOSE  Open a command (emulate popen)
.RETURNS  The file handle
.REMARKS  Use a vfork (popen can't work in ftpd mode)
	pid of child process stored as static.
-----------------*/
  char **argv;	/* IN: Function & Arguments (terminated by 0) */
  int  mode;	/* IN: 0=Read, 1=Write */
{
  static int fifo[2], pid, mod1;
  char *p ;

    if (!argv)  return(mode);	/* No command ! */
    if (!*argv) return(mode);	/* No command ! */
#ifdef __MSDOS__ 	/* For tests */
    fprintf(stderr, "++++CALL open_pipe(\"%s\", %d)\n", *argv, mode);
    return (open(*argv, mode));
#else
    mod1 = mode^1;	/* Transforms 0 into 1 and vice-versa */
    if (pipe(fifo)) { perror(*argv); return(-1); }

#if DEBUG
    fprintf(stderr, "....open_pipe(%s,%d), fifo=(%d,%d), mode=%d, mod1=%d\n",
	*argv, mode, fifo[0], fifo[1], mode, mod1);
#endif

    child = 0 ;
    pid = fork();	/* NOT fvork!! */
    if (pid < 0) {
	perror(*argv);
	fprintf(stderr, "****vfork to exec '%s' with mode=%d failed\n",
	  *argv, mode);
	return(-1);
    }
    if (pid) {		/* Parent just closes the useless side of pipe */
#if DEBUG
	fprintf(stderr, "....I'm the parent!\n");
	fprintf(stderr, "....open_pipe(%s,%d): father created pid=%d\n",
		*argv, mode, pid);
	fprintf(stderr, "    fifo=(%d,%d), mode=%d, mod1=%d\n",
	    	fifo[0], fifo[1], mode, mod1);
	fprintf(stderr, "    fifo=(%d,%d) fifo[mode]=%d, fifo[mod1]=%d\n",
		fifo[0], fifo[1], fifo[mode], fifo[mod1]);
#endif
	child = pid;
	signal(SIGCHLD, SIG_IGN);
	close(fifo[mod1]);
#if DEBUG	/* DEBUG */
	for (i=0; i<=2; i++) fprintf(stderr,
	   "....fzopen-father: fstat(%d)=%d\n", i, fstat(i, &bstat)) ;
#endif	/* DEBUG */
	return(fifo[mode]);
    }
	/* Here the child process has to execute the command */
    close(mod1); if (dup(fifo[mod1])<0) perror("#***open_pipe");
    close(fifo[0]); close(fifo[1]);
#if DEBUG	/* DEBUG */
	for (i=0; i<=2; i++) fprintf(stderr,
	   "....fzopen--child: fstat(%d)=%d\n", i, fstat(i, &bstat)) ;
#endif	/* DEBUG */
    p = getenv("PATH") ;
    if (!p) p = "(nil)" ;
    if (verbop) { char **a;
	fprintf(stderr, "....EXEC");
	for (a=argv; *a; a++) fprintf(stderr, " %s", *a);
	fprintf(stderr, "\n    PATH=%s\n", p);
	fflush(stderr);
    }
    execvp(*argv, argv);	/* Replace child by executable */
    if (getuid() == 0) syslog(LOG_ERR, "***Can't exec %s, PATH=%s\n",
	*argv, p) ;
    else { perror(*argv); fprintf(stderr, "****PATH=%s\n", p) ; }
    return(-1);
#endif
}

static int my_open(char *name, char *ext)
/*++++++++++++++++
.PURPOSE  Effectively open the file (via a pipe)
.RETURNS  ---
.REMARKS  Choose between gzip and bzip2
-----------------*/
{
  char *buf;
  int fd;
    child = 0;		/* Default chid pid */
    zcat[0] = "gzip"; zcat[1] = "-cd";
    if (ext[0] == '.') switch(ext[1]) {
	case 'b': 	/* Command: bzip2 -cd file... */
	    zcat[0] = "bzip2"; 
	    break;
	case 'f': 	/* Command: funpack -S file... */
	    zcat[0] = "funpack"; zcat[1] = "-S"; 
	    break;
    }
    if ((buf = zcat[zcatfile])) free(buf);
    buf = malloc(1+strlen(name)+strlen(ext));
    strcpy(buf, name) ;
    strcat(buf, ext) ;
    zcat[zcatfile] = strdup(buf);
    fd = open_pipe(zcat, 0);		/* Open for READ */
    return(fd);
}


static void on_hup(sig) int sig ;
{
  static int count = 0;
  int st = 0;
    inSig |= 1; count++;
    if (verbop) { 
        fprintf(stderr, "\n++++fcat: stop on signal#%d", sig);
	if (sig == SIGPIPE) fprintf(stderr, "(SIGPIPE)");
	if ((getpid() == root) && child) {	/* V2.0: propagate signal */
	    fprintf(stderr, " (sent to child)\n");
	    fflush(stderr);
	}
    }
    if ((getpid() == root) && child) {
        signal(sig, on_hup);
	if (kill(child, sig /*SIGHUP*/))  /* An error in signal */
	    exit(1);
	while (waitpid(child, &st, WNOHANG) == 0) {  /* Not Yet Finished? */
	    if (++count>9) break;
	    sleep(1);
	}
	st = WEXITSTATUS(st);
	child = 0;
	if (verbop) fprintf(stderr, "....Status of child: 0x%04X\n", st);
    }
    if (count>9) st = 9;
    if (st) exit(st);
}

/*==================================================================
		Internal Routines
 *==================================================================*/
static int value(char *text)
/*++++++++++++++++
.PURPOSE  Interpret a value eventually followed by 'k', 'M', etc
.RETURNS  The scaled value (in bytes) / -1 if error
-----------------*/
{
  int val; 
    if (!text) return(-1);
    val = atoi(text);
    if (val == 0) val = 1;
    while (isdigit(*text)) text++;
    switch(*text) {
      case 'c':
	/* NOBREAK */
      case 0: 
        break;
      case 'b':
        val *= 512;
        break;
      case 'f': case 'F':	/* FITS blocks */
        val *= 2880;
        break;
      case 'k': case 'K':
        val *= 1024;
        break;
      case 'm': case 'M':
        val *= 1024*1024;
        break;
      default:
        val = -1;
    }
    return(val);
}

static int read_full(int fd, char *buf, int buf_size)
/*++++++++++++++++
.PURPOSE  Read completely a buffer
.RETURNS  Number of bytes (if < buf_size, end reached) / -1 if error
-----------------*/
{
  char *p, *e;
  int stat;
    p = buf; e = p+buf_size; 
    if (verbop) fprintf(stderr, "....read_full(%d) =>", (int)(e-p));
    while (p<e) {
	stat = read(fd, p, e-p);
	if (verbop) {
	    if (inSig) fprintf(stderr, " +++(read_full interrupted)+++");
	    fprintf(stderr, " %4d", stat);
	    fflush(stderr);
	}
	if (stat < 0) {
	    if (verbop) fprintf(stderr, "\n");
	    perror("****read_full");
	    return(-1);
	}
	if (stat == 0)	/* EOF reached */
	    break;
	p += stat;
    }
    if (verbop) fprintf(stderr, "\n");
    return(p-buf);
}

static int write_full(int fd, char *buf, int bytes)
/*++++++++++++++++
.PURPOSE  Write completely a buffer
.RETURNS  bytes / -1
-----------------*/
{
  char *p, *e;
  int stat;
    p = buf; e = p+bytes; 
    if (verbop) fprintf(stderr, "....write_ful(%d) =>", (int)(e-p));
    while (p<e) {
	stat = write(fd, p, e-p);
	if (verbop) {
	    if (inSig) fprintf(stderr, " +++(write_ful interrupted)+++");
	    fprintf(stderr, " %4d", stat);
	    fflush(stderr);
	}
	if (stat < 0) {
	    if (verbop) fprintf(stderr, "\n");
	    if ((errno == EINTR) || (errno == EPIPE)) break;
	    perror("****write_full");
	    return(-1);
	}
	if (stat == 0)	/* EOF reached */
	    return(-1);
	p += stat;
    }
    if (verbop) fprintf(stderr, "\n");
    return(p-buf);
}

static void message (char *name, int _end)
/*++++++++++++++++
.PURPOSE  Printer the header (-m option)
.RETURNS  ---
-----------------*/
{
static char pcrule[] = "\
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" ;
  char *p;
    if (_end) { /* printf("%%\\end{Message} %%%s\n", name) ; */ return; }
    for (p=name; isdigit(*p); p++) ;
    if (*p) printf("\n#==> %s <==#\n", name) ;
    else printf("\n%s\n%%\tMessage %s\n%s\n%%\\begin{Message} %%%s\n",
	pcrule, name, pcrule, name) ;
    fflush(stdout) ;
}

static char *which_ext(char *name)
/*++++++++++++++++
.PURPOSE  Find the list of files matching a name
.RETURNS  The extension, either in zext list, or as the highest
		number when a file was split into pieces.
	  or "/" for all files in a directory, files saved in lnames .
-----------------*/
{
  static char numext[12] ;
  LNAME *pname, *lname ;
  DIR *dirp; struct dirent *dire;
  struct stat buf;
  char dirname[LEN_CWD] ;
  char *p, *fname, *match;
  int flen, dlen, i, val ;
  short split[2] ;	/* Min and Max when Large File cut in Pieces */

    /* Separate the Directory and File Name */
    strncpy(dirname, name, sizeof(dirname)-1) ;
    dirname[sizeof(dirname)-1] = 0 ;
    for (p=dirname+strlen(dirname) - 1; (p >= dirname) && (*p != '/'); p--)  ;
    if (p == dirname)     dirname[0] = '/', dirname[1] = 0, fname = name ;
    else if (p < dirname) dirname[0] = '.', dirname[1] = 0, fname = name ;
    else {
	++p; 	/* Keep the / */
	fname = name + (p-dirname) ;
	*(p++) = '.' ;		/* Keep "/." for Links */
	*(p++) = 0;
    }

    if (verbop) fprintf(stderr,
	"\n....Looking for files: %s.* in directory: %s\n", fname, dirname) ;

    if (lstat(dirname, &buf) < 0) {
      	if (!optq) { fprintf(stderr, "****fcat ") ; perror(dirname) ; }
        return((char *)0) ;
    }

    if (!S_ISDIR(buf.st_mode))
    	return((char *)0) ;

    dirp = opendir(dirname) ;
    if (!dirp) {
	if (!optq) { 
	    perror("\n++++");
	    fprintf(stderr, "++++Can't opendir '%s'\n", name);
	}
	return((char *)0);
    }

    /* Read the Directory and find names compatible */
    flen = strlen(fname) ;
    dlen = strlen(dirname) ;
    match = (char *)0 ;
    lname = pname = (LNAME *)0 ;
    split[0] = 9999 ; split[1] = -1 ;
    while ((dire = readdir(dirp))) {
	if (dire->d_name[0] == '.') continue ;		/* Hidden file   */
	if ((fname[0] == '*') && (fname[1] == 0)) {	/* get ALL files */
	    pname = lname ;
	    lname = (LNAME *)malloc(sizeof(LNAME) + dlen
		  + strlen(dire->d_name)) ;
	    if (pname) pname->next = lname ; else lnames = lname ;
	    lname->next = (LNAME *)0 ;
	    strcpy(lname->name, dirname) ;
	    lname->name[dlen] = '/' ;
	    strcpy(lname->name + dlen + 1, dire->d_name) ;
	    /* Remove the Compress Extensions */
	    p = lname->name + strlen(lname->name) ;
	    for (i=0; i<ITEMS(zext); i++) {
		if (strcmp(p-strlen(zext[i]), zext[i])) continue ;
		p -= strlen(zext[i]) ;
		*p = 0 ;
		break ;
	    }
	    match = "/" ;
	    continue ;
	}
	if (strncmp(dire->d_name, fname, flen))	/* Mismatching name */
	    continue  ;
	p = dire->d_name + flen ;
	if (p[0] != '.') continue ;
        if (verbop) fprintf(stderr,
	    "....filename examined: %s\n", dire->d_name) ;
	for (i=0; i<ITEMS(zext) && strcmp(p, zext[i]); i++) ;
	if (i<ITEMS(zext)) { match = zext[i]; break ; }
	/* Try the numeric possibilities */
	if (!isdigit(p[1])) continue ;
	i = 0 ;
	numext[i++] = *(p++) ;
	for (val=0; (i<5) && isdigit(*p); p++)
	    numext[i++] = *p, val = val*10 + (*p - '0') ;
	numext[i] = 0 ;
	if (*p) {
	    if (*p != '.') continue ;
	    for (i=0; i<ITEMS(zext) && strcmp(p, zext[i]); i++) ;
	    if (i >= ITEMS(zext)) continue ;
	}
	if (val < split[0])  split[0] = val ;
	if (val > split[1])  split[1] = val ;
    }
    closedir(dirp) ;

    /* Return the extension found */
    if (match) return(match) ;

    if (split[0] < split[1]) {
	val = split[1] ;
    	for (i=strlen(numext)-1; i>0; i--)
	    numext[i] = '0' + (val%10), val /= 10 ;
	return(numext) ;
    }
    return((char *)0) ;
}

static int list_files(char *name)
/*++++++++++++++++
.PURPOSE  Print the names of the files making up the catalog
.RETURNS  Number of files
-----------------*/
{
  LNAME *lname, *nname ;
  int i, status, fd, fno, fmax, nd, val, iroot ;
  char buf[BUFSIZ], anum[12] ;
  struct stat sbuf;
  char *ext ;

    status  = stat(name, &sbuf) ;
    if (status == 0) {
        printf("%s", name) ;
    	/* DIRECTORY is listed with a trailing / */
	iroot = strlen(name) -1;
	if (S_ISDIR(sbuf.st_mode) && (name[iroot] != '/')) 
	    putchar('/');
	putchar('\n');
	return(1) ;
    }

    /* Not found, maybe the file is given a compressed extension (V2.2) */
    for (i=strlen(name)-1; (i>0) && (name[i] != '.'); i--) ; iroot = i;
    if (iroot>0) {	/* Mod. V2.41 */
	for (i=0; i<ITEMS(zext); i++) {
	    if(strcmp(name+iroot, zext[i]) == 0) break;
        }
        if (i<ITEMS(zext)) {	/* Yes file exists */
            fwrite(name, 1, iroot, stdout);
	    putchar('\n');
	    return(1);
        }
    }

    ext = which_ext(name) ;
    if (!ext)  {
	if (!optq) { fprintf(stderr, "****fcat: ") ; perror(name) ; }
	exit(1) ;
    }

    memset(buf, 0, sizeof(buf)) ;
    strncpy(buf, name, sizeof(buf) - strlen(ext) - 5) ;
    iroot = strlen(buf) ;
    if (ext[0] == '/') {	/* List of Files in a Subdirectory */
	for (lname = lnames, fmax=0; lname; lname = nname) {
	    nname = lname->next ;
	    status = list_files(lname->name) ;
	    free(lname);
	    fmax+= status ;
	}
	lnames = (LNAME *)0 ;
	return(fmax) ;
    }
    if (isdigit(ext[1])) {	/* File split into pieces. Recurvisely call */
	strcpy(anum, ext) ;
	for (nd=0, fmax=0, ++ext; isdigit(*ext); ext++)
	    fmax = (fmax*10) + (*ext - '0'), nd++ ;
	for (fno=0; fno <= fmax; fno++) {
	    for (fd = nd, val = fno; fd>0; fd--, val /= 10)
		anum[fd] = '0' + (val%10) ;
	    strcpy(buf+iroot, anum) ;
	    status = list_files(buf) ;
    	}
	return(fmax+1) ;
    }
    else {		/* Start the Uncompress Utility */
	printf("%s%s\n", buf, ext) ;
	return(1) ;
    }
}


static int exec_cat(char *name)
/*++++++++++++++++
.PURPOSE  Print the contents of a file; unzipped if necessary
.RETURNS  0 = OK
-----------------*/
{
  LNAME *lname, *nname ;
  int fd, fno, fmax, nd, val, iroot, pid, stat, buf_size=0, done, i;
  char *buf=0, *ext, anum[12] ;
  int records, ignore, bs1;

    /* Allocate buffer: size is a multiple of specified bs */
    bs1 = bs ? bs : 1;
    buf_size = BUFSIZ/bs1;
    if (!buf_size) buf_size = 1;
    buf_size *= bs1;
    buf = malloc(buf_size);

    if (verbop) fprintf(stderr, 
	    "....fcat: reading %s (bufsize=%d):", name, buf_size) ;
    pid = 0 ;
    if (strcmp(name, "-") == 0) fd = 0;
    else fd = open(name, 0) ;

    if ((fd < 0) && (errno == ENOENT)) {
        ext = which_ext(name) ;
	if (!ext)  {
	    if (verbop) fprintf(stderr, "\n") ;
	    if (!optq) { fprintf(stderr, "****fcat: ") ; perror(name) ; }
	    exit(1) ;
	}
	memset(buf, 0, buf_size) ;
	strncpy(buf, name, buf_size - strlen(ext) - 5) ;
	iroot = strlen(buf) ;
        if (ext[0] == '/') {		/* List of Files in a Subdirectory */
	    for (lname = lnames; lname; lname = nname) {
	        nname = lname->next ;
	        stat = exec_cat(lname->name) ;
		if (stat) exit(stat) ;
	        free(lname);
	    }
	    lnames = (LNAME *)0 ;
	    free(buf);
	    return(0);
        }
	if (isdigit(ext[1])) {	/* File split into pieces. Recurvisely call */
	    strcpy(anum, ext) ;
	    for (nd=0, fmax=0, ++ext; isdigit(*ext); ext++)
	        fmax = (fmax*10) + (*ext - '0'), nd++ ;
	    if (verbop) fprintf(stderr, " (in %d pieces)\n", fmax+1) ;
	    for (fno=0; fno <= fmax; fno++) {
	        for (fd = nd, val = fno; fd>0; fd--, val /= 10)
		    anum[fd] = '0' + (val%10) ;
		strcpy(buf+iroot, anum) ;
		stat = exec_cat(buf) ;
		if (stat) exit(stat) ;
	    }
	    free(buf);
	    return(0);
	}
	else {		/* Start the Uncompress Utility */
	    fd = my_open(name, ext);		/* Open via decompressor */
	    pid = child ;			/* For Closing   */
	}
    }

    if (verbop) fprintf(stderr, "....fd=%d\n", fd);

    /* Error Check */

    if (fd < 0) {
        if (verbop) fprintf(stderr, "\n") ;
        if (!optq) { fprintf(stderr, "****fcat: ") ; perror(name) ; }
	if (pid) waitpid(pid, &stat, 0) ;
	exit(1) ;
    }

    /* Display the file */
    if (verbop) fprintf(stderr, "\n") ;
    records = ignore = 0 ;
    if (ohead) message(name, 0) ;
    fmax = 0; inSig = 0;
    siginterrupt(SIGPIPE, 1);
    signal(SIGPIPE, on_hup);

    /* Skip value: TODO */
    if (skip) { struct stat bstat;
	if (fstat(fd, &bstat) < 0) 
	    perror("****stat");
	else {
	    if (verbop) fprintf(stderr, 
		"....skip_full(%dx%d): %sregular file\n", skip, bs1,
		S_ISREG(bstat.st_mode) ? "" : "non-");
	    /* Can seek only when the blocksize if known... */
	    if (S_ISREG(bstat.st_mode) && bs)
		 lseek(fd, bs*skip, SEEK_SET);
	    else ignore = skip;
	}
    }

    /* Display what's required */
    done = 0;
    stat = 0;
    while (done == 0) {
	stat = read_full(fd, buf, buf_size);
	if (stat <= 0) break;
	val = stat/bs1;		/* Number of records */
	if (ignore) {
	    if (bs==0) { 	/* Count the newlines */
		for (i=0; (i<stat) && ignore; i++) {
		    if (buf[i] == '\n') ignore--;
		}
		if (i>=stat) continue;
		stat -= i;		/* New value of number of bytes */
		memmove(buf, buf+i, stat);
	    }
	    else if (val <= ignore) { ignore -= val; continue; } 
	    /* Partial ignore: move bytes in the buffer */
	    else {
		stat -= (ignore*bs1);	/* New value of number of bytes */
	        memmove(buf, buf+(ignore*bs1), stat);
	    }
	    ignore = 0;			/* Skipped records terminated   */
	}
	if (bs == 0)  		/* Count the newlines */
	for (i=0; i<stat; i++) {
	    if (buf[i] != '\n') continue;
	    records++;
	    if (count && (records >= count)) { 
	        stat = i+1; 
	        done = 1; 
	        break;
	    }
	}
	else {			/* Use blocksize */
	    records += stat/bs1;
	    if (count && (records >= count)) {
		i = records - count;	/* Records in excess */
		stat -= bs1 * i;
		records -= i;
		done = 1;
	    }
	}
	stat = write_full(1, buf, stat);
	if (stat < 0) break;
    }
    if (verbop && inSig) fprintf(stderr, "++++(was interrupted)\n");

    if (verbop) {
	fprintf(stderr, "----exec_cat(%s): transfer=%d", name, records);
	if (bs1>1) fprintf(stderr, "x%d", bs1);
	if (inSig) fprintf(stderr, " +++(was interrupted)+++");
	fprintf(stderr, "\n");
	fflush(stderr);
    }
    if (ohead) message(name, 1) ;
    close(fd) ;

#if 0		/* SIGCHILD ignored ... */
    if (pid) {
	if (verbop) fprintf(stderr, "....waitpid(%d) => ", pid);
        waitpid(pid, &stat, 0) ;
	if (verbop) fprintf(stderr, "%d\n", stat);
    }
#endif

    if (stat < 0) {
        if (!optq) { fprintf(stderr, "****fcat: ") ; perror(name) ; }
	exit(1) ;
    }

    if (buf) free(buf);
    return (0);
}


/*==================================================================
		Program
 *==================================================================*/
int main (int argc, char **argv)
{
  struct stat sbuf;
  int err, nc, status, len, i, *aint;
  char *ext, *p, *e, get_max;

    nc  = 0;
    err = 0 ;	/* Returned Code */
    get_max = 0;
    root = getpid();

	/* Check the Arguments */

    if (--argc < 1) {
        fprintf(stderr, "%s", usage);
	exit(1) ;
    }
    aint = (int *)0;		/* Value to get */

    while ((--argc >= 0) && (err == 0)) {
	p = *++argv;
	/* V2.0: accept arguments as count= skip= bs= */
	if ((e = strchr(p, '=')) && isdigit(e[1])) { 
	    if     (starts_by(p, "count=")) aint = &count;
	    else if (starts_by(p, "bs="))   aint = &bs;
	    else if (starts_by(p, "skip=")) aint = &skip;
	    if (aint) {		/* Argument written ala dd */
		*aint = value(e+1);
		if (*aint < 0) {
		    fprintf(stderr, "****Bad option: %s\n%s", *argv, usage);
		    exit(1);
		}
		aint = (int *)0; 
		continue;
	    }
	}

	if (*p == '-') for (++p; *p; p++) switch (*p) {
	  case 'h': 	/* Help or Header ? */
	    if (argc || argv[1]) {	/* Not help... */
	        ohead = 1;
	        continue;
	    }
	    /* NO BREAK */
	      case '?':
	    printf("%s%s", usage, help) ;
	    exit(0) ;
	  case 'v' :
	    if (strncmp(p, "ver", 3) == 0) {
	        printf("fcat (CDS) Version %s\n", VERSION);
	        exit(0);
	    }
	    verbop = 1 ;
	    continue ;
	  case 'f' :
	    optf = 1 ;
	    continue ;
	  case '1' :
	    opt1 = 1 ;
	    continue ;
	  case 'q' :
	    optq = 1 ;
	    continue ;
	  case 'm' :
	    ohead = 1 ;
	    continue ;
	  case 'b':		/* Block-size */
	    if (p[1])     bs = value(p+1);
	    else  --argc, bs = value(*++argv);
	    if (bs<0) goto bad_option;
	    continue;
	  case 'c':		/* Count */
	  case 's':		/* Stop after */
	    if (isdigit(p[1])) count = value(p+1);
	    else       --argc, count = value(*++argv);
	    if (count<0) goto bad_option;
	    continue;
	  case 'i':		/* Ignore */
	    if (isdigit(p[1])) skip = value(p+1);
	    else       --argc, skip = value(*++argv);
	    if (skip<0) goto bad_option;
	    continue;
	  bad_option:
	  default:
	    fprintf(stderr, "****Bad option: '%c' in %s\n%s", 
		*p, *argv, usage);
	    exit(1) ;
	} 
	else {
	    nc++ ;
	    if (opt1) {			/* List the name of the files	*/
	        status = list_files(p) ;
	        if (status == 0) {	/* No File was printed....	*/
		    err++ ;
		    if (!optq) 
			fprintf(stderr, "****fcat: No such file: %s\n", p) ;
		}
	    }
	    else if (optf) {		/* Just verify the file exists! */
	        if (stat(p, &sbuf) < 0) {
		    if (which_ext(p)) ;	/* valid extension  ??	*/
		    else err = 1 ;
		}
		if (err) break ;
	    }
	    else {
		/* Remove the Extension */
		e = p + strlen(p) ;
		for (i=0; i<ITEMS(zext); i++) {
		    ext = zext[i] ;
		    len = strlen(ext) ;
		    if (strcmp(e-len, ext) == 0)
			*(e-len) = 0 ;
		}
		exec_cat(p) ;
	    }
	    continue ;
	}
    }

    if (nc == 0) { 	/* Cat the Standard Input */
	if (optf || opt1) err = 1 ;
        else exec_cat("-");
    }
    if (child) on_hup(SIGTERM); /*wait(&i);*/

    exit(err) ;
}
