/*++++++++++++++
.IDENTIFICATION filutil.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   15-May-1991 (anafile)
.VERSION  4.0   28-Feb-1994 New package
.VERSION  4.1   21-Dec-1994: in daemon env., no popen
.VERSION  4.2   26-Mar-1997: filename function, ext NULL
.VERSION  4.3   17-Jul-1999: Corrected bug in open_pipe
.VERSION  4.4   05-Jan-2000: Error logged uin syslog when run by root
.VERSION  4.5   05-Feb-2000: fzopen uses the fcat program
			to allow reading files split in pieces
.VERSION  4.6   29-Jul-2005: fzopen decompresses for compressed file.
.VERSION  4.7   26-Nov-2005: fxopen / fxclose
.VERSION  4.8   20-Feb-2006: load_file: error messages
.VERSION  4.9   02-Jun-2007: Cosmetics (-Wall)
.VERSION  5.0   19-Dec-2007: also http files (use wwwget)
.VERSION  5.1   28-Apr-2008: also http files (use wwwget or wget)
.VERSION  5.2   19-Aug-2008: added launch_cleaner
.VERSION  5.21  22-Oct-2008: kill the sub-process
.VERSION  5.22  30-Dec-2008: Pb if a mixture of compressed/uncompressed/http
.VERSION  5.23  15-Apr-2010: fzclose uses SIGTERM (instead of SIGHUP)
.VERSION  5.24  12-Sep-2010: load_stream
.VERSION  5.3   04-Aug-2011: Accept files as pipe if starts by |
.VERSION  5.4   12-May-2012: Avoid the SIG_IGN
.VERSION  5.41  09-Jul-2013: Accept .fz (FITS rice compression)
	fcat is a program which lists to stdout a file, even when it's
	compressed, split into parts, etc.
	fcat has a -f option which just checks for the file existence.
.COMMENTS       Small C Utilities for File Management
---------------*/

/* Problems with Fork / vfork.
   Better to use vfork which is faster, but doesn't exist on some machines...
*/
#ifdef NOvfork
#define vfork fork
#endif

#ifndef MAX_PIPES
#define MAX_PIPES  64	/* via open_pipe */
#endif

/* Which program for http://... files ?  */
#ifndef WWWGET
#define WWWGET  "NO_HTTP"
#endif

#ifndef DEBUG
#define DEBUG	0
#endif

#include <filutil.h>	/* Standard definitions	*/
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>	/* File Control		*/
#include <time.h>
#ifndef O_BINARY
#define O_BINARY 0
#endif
#ifdef __MSDOS__
#include <sys\stat.h>
#else
#include <sys/stat.h>
#include <sys/wait.h>	/* wait waitpid 	*/
#include <sys/types.h>
#include <dirent.h>	/* Read the Directory 	*/
#include <signal.h>	/* To release the Zombies */
#include <syslog.h>	/* For Root ...	*/
#endif
#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])
#define isquote(c)	(((c) == '\'') || ((c) == '"'))

#if 0
static char *zcat[] = {	/* Command + Options to decompress a file */
	"gzip", "-c", "-d",  (char *)0, (char *)0
};
#define zcatfile	(ITEMS(zcat)-2)
#endif

static char *zext[] = { 	/* Extensions & Programs */
	".Z",   /* Standard Unix compress */
	".z",   /* GNU zip */
	".gz",
	".bz",  /* bzip2   */
	".bz2", /* bzip2   */
	".fz",  /* FITS rice (http://heasarc.gsfc.nasa.gov/fitsio/fpack/) */
};
/* Table with special extensions */
static char *special_ext[] = {
    ".awk", "gawk -f ",
    ".sed", "sed -f ",
    ".sh",  "sh ",
    ".bash","bash ",
} ;
static char root_dir[] = { dirsep, 0 } ;
static char *wwwget = WWWGET;

/*==================================================================
		Pipe utilities
 *==================================================================*/

#ifdef __MSDOS__ 	/* For tests */
int wait_pipe(opt)	/* Useless in MS-DOS */
   { return(0); }
#else			/* == Standard Unix */

static unsigned char opipes[1024];	/* fileno --> child+header... */
static int children[MAX_PIPES];		/* pid of child, -1=reg, -2   */
static int chilstat[MAX_PIPES];		/* Status of child from wait  */
static char *header[MAX_PIPES];		/* File name, or
					   HTTP-header for http://... */
/* Header starts by "#\t+seek_offset for gz-index */

static void on_death(int dummy)
/*++++++++++++++++
.PURPOSE  Trapping Routine to remove Zombies
.RETURNS  ---
.REMARKS  Read child's status and remove the pid in the table
-----------------*/
{
  int st, pid, i;
#if DEBUG
    fprintf(stderr, "....on_death() called; wait");
#endif
    pid = wait(&st);
#if DEBUG
    fprintf(stderr, " pid=%d, child status=%d\n", pid, st);
#endif
    if (st) fprintf(stderr, "++++Child ended with status: %d\n", st);
    for (i=0; i<ITEMS(children);  i++) {
	if (children[i] == pid) {
	    chilstat[i] = st ;
	    break;
	}
    }
}

static int add_info(FILE *f, char *name, int type)
/*++++++++++++++++
.PURPOSE  Add the details in the opipes + header
.RETURNS  For http (type=80): the status. Otherwise 0=OK, -1=error
.REMARKS  type=0: standard file ; type=80(P): HTTP
          type=*: executable file
	  type=|: shell pipe
-----------------*/
{
  int i, status, len, allo, used, fh;
  char *a, *p;
  struct stat bstat;
  char buf[512];
    if (!f) 	/* NULL pointer */
	return(-1);
    fh = fileno(f);
    if ((fh<0) || (fh >= sizeof(opipes))) {
	fprintf(stderr, "#***filutil: impossible file number: %d\n", fh);
	return(-1);
    }
    for (i=0; i<ITEMS(header) && header[i]; i++) ;
    if (i>=ITEMS(header)) {
	fprintf(stderr, "#***filutil: too many opened files (%d)\n", i);
	return(-1);
    }
    opipes[fh] = i;
    switch(type) {
      case 0:		/* A file -- save its name */
	type = '?';	/* File of unknown type... */
	if (fstat(fh, &bstat) == 0) {
	    /* '=' as first byte means "regular file" */
	    if (S_ISREG(bstat.st_mode)) type = '=';
	}
      default:
	buf[0] = 0;	/* type is now given... */
	if (name[0] != dirsep) {
	    if (getcwd(buf, sizeof(buf)-1))
		strcat(buf, root_dir);
	    else buf[0] = 0;
	}
	a = malloc(strlen(name)+strlen(buf)+8);	/* Add .gzi if necessary */
	header[i] = a;
	a[0] = type;
	strcpy(a+1, buf); a += strlen(a);
	strcpy(a, name);
	return(0);
      case 80:		/* Get HTTP header */
	status = 0;
	a = (char *)0;
	allo = used = 0;
	while (1) {
	    if (!fgets(buf, sizeof(buf), f)) 	/* Can't read first line */
		buf[0] = 0;
	    if (strncmp(buf, "HTTP", 4) != 0) {	/* Not a HTTP answer ??? */
		if (a) free(a);
		return(-1);
	    }
	    for (p=buf; isgraph(*p); p++) ; while (isspace(*p)) p++;
	    status = atoi(p);
	    /* Get the full header */
	    while (1) {
		len = strlen(buf);
		if ((used+len)>=allo) {
		    allo += (len|1023) +1;
		    if (a) a = realloc(a, allo);
		    else   a =  malloc(allo);
		}
		strcpy(a+used, buf);
		used += len;
		if (fgets(buf, sizeof(buf), f)) {
		    if ((buf[0] == '\r') && (buf[1] == '\n'))	/* End header */
			break;
		}
		else break;
	    }
	    if ((status >= 300) && (status < 400)) 		/* Redirect */
		used = 0;
	    else break;
	}
	header[i] = a;		/* Save header */
	return(status);
    }
    return(-1);
}

int wait_pipe(int opt /* For future use */)
/*++++++++++++++++
.PURPOSE  Wait until all opened pipes are closed
.RETURNS  0 / -1
.REMARKS  Just performs a wait until all pipes opened via open_pipe are closed
-----------------*/
{
  int i;
#if DEBUG	/* List the used slots */
    fprintf(stderr, "....wait_pipe has slots:");
    for (i=0; i<ITEMS(children); i++) if (children[i])
        fprintf(stderr, " %d", children[i]);
    fprintf(stderr, "\n");
#endif
    for (i=0; i<ITEMS(children); i++) {
	if (children[i]) on_death(0);
    }
    return(0) ;
}
#endif

int open_pipe(char **argv /* Function & Args */, int mode /*0=read, 1=wr*/)
/*++++++++++++++++
.PURPOSE  Open a command (emulate popen)
.RETURNS  The file handle
.REMARKS  Use a vfork (popen can't work in ftpd mode)
-----------------*/
{
  int fifo[2], i, pid, mod1;
  char *p ;

    if (!argv)  return(mode);	/* No command ! */
    if (!*argv) return(mode);	/* No command ! */
#ifdef __MSDOS__ 	/* For tests */
    fprintf(stderr, "++++CALL open_pipe(\"%s\", %d)\n", *argv, mode);
    return (open(*argv, mode));
#else
    mod1 = mode^1;	/* Transforms 0 into 1 and vice-versa */
    if (pipe(fifo)) { perror(*argv); return(-1); }
    pid = vfork();
    if (pid < 0) {
	perror(*argv);
	fprintf(stderr, "****vfork to exec '%s' with mode=%d failed\n",
	  *argv, mode);
	return(-1);
    }
    if (pid) {	/* Father just closes the useless side of pipe */
			/* Find a slot in table of children   */
        for (i=0; children[i] && (i < ITEMS(children)); i++ ) ;
#if DEBUG
	fprintf(stderr, "....open_pipe(%s) slot#%d, created pid=%d\n",
		*argv, i, pid);
#endif
	if (i < ITEMS(children)) {
	    children[i] = pid;
	    chilstat[i] = -1 ;
	}
	else fprintf(stderr, "****Too many children (%d), pid=%d\n", i, pid) ;
	signal(SIGCHLD, on_death /*SIG_IGN*/);
	close(fifo[mod1]);
#if 0	/* DEBUG */
	for (i=0; i<=2; i++) fprintf(stderr,
	   "....fzopen-father: fstat(%d)=%d\n", i, fstat(i, &bstat)) ;
#endif	/* DEBUG */
	return(fifo[mode]);
    }
	/* Here the child process has to execute the command */
    close(mod1); if (dup(fifo[mod1])<0) perror("#***dup");
    close(fifo[0]); close(fifo[1]);
#if 0	/* DEBUG */
	for (i=0; i<=2; i++) fprintf(stderr,
	   "....fzopen--child: fstat(%d)=%d\n", i, fstat(i, &bstat)) ;
#endif	/* DEBUG */
    execvp(*argv, argv);	/* Replace child by executable */
    p = getenv("PATH") ;
    if (!p) p = "(nil)" ;
    if (getuid() == 0) syslog(LOG_ERR, "***Can't exec %s, PATH=%s\n",
	*argv, p) ;
    else { perror(*argv); fprintf(stderr, "****PATH=%s\n", p) ; }
    return(-1);
#endif
}

/*==================================================================
		File utilities
 *==================================================================*/
int is_dir(char *name)
/*++++++++++++++++
.PURPOSE  Check whether a file represents a directory
.RETURNS  0 (not a directory) / 1 (is a directory)
-----------------*/
{
  struct stat bstat;
    if (stat(name, &bstat)) 	/* Error */
    	return(0);
    return(S_ISDIR(bstat.st_mode));
}

int fprintq(FILE *file, char *string)
/*++++++++++++++++
.PURPOSE  Print a quoted string using C-like standards
.RETURNS  0
-----------------*/
{
  char *p;
    fputc('"', file);
    for (p=string; *p; p++) {
    	if (isprint(*p)) {
    	    if (*p == '"') fputc('\\', file);
    	    fputc(*p, file);
    	}
    	else fprintf(file, "\\%03o", (*p)&0xff);
    }
    fputc('"', file);
    return(0);
}

int copy_file(name, out)
/*++++++++++++++++
.PURPOSE  Copy a file to an opened output
.RETURNS  0 (OK) / -1 (errors)
-----------------*/
  char *name;	/* IN: Name of the file to load */
  int out;	/* OUT: File opened for output  */
{
  int  fno, n;
  char *buffer;

	/* Load file into Memory */
    fno = open(name, O_BINARY);
    if (fno <= 0) {
    	perror(name);
    	return (-1);
    }

    buffer = malloc(8192);
    while(1) {
	n = read(fno, buffer, 8192);
	if (n <= 0) break;
	if (write(out, buffer, n) < 0) {
	    perror("Output file");
	    return (-1);
	}
    }
    close(fno);
    free(buffer);
    return(0);
}

char *load_stream(FILE *fp)
/*++++++++++++++++
.PURPOSE  Load the contents of an opened file into memory
.RETURNS  The loaded text terminated by NULL / NULL if error
-----------------*/
{
  char buffer[BUFSIZ];
  char *abuf=0, *cbuf, *ebuf;
  int n, o, bytes, nallo=0;

    /* Read first in a local buffer ... */
    cbuf = buffer, ebuf = cbuf + sizeof(buffer)-1;
    while ((bytes = fread(cbuf, 1, (ebuf-cbuf), fp)) > 0) {
	cbuf += bytes; *cbuf = 0;
	if ((ebuf-cbuf)<16) break;
    }

    /* Allocate the first buffer */
    if (cbuf > buffer) {	/* Something was read... */
	o = (cbuf-buffer);	/* #bytes filled */
	n = bytes ? 2*sizeof(buffer) : o+1;
	abuf = malloc(n);
	cbuf = abuf + o;
	ebuf = abuf + (n-1);
	memcpy(abuf, buffer, o);
	if (bytes) while ((bytes = fread(cbuf, 1, (ebuf-cbuf), fp))>0) {
	    cbuf += bytes;
	    if ((ebuf-cbuf)<16) { nallo++;
		o = (cbuf-abuf);
		n = (ebuf-abuf) + BUFSIZ; /* Expand the buffer */
		abuf = realloc(abuf, n);
		if (!abuf) {
		    fprintf(stderr, "#***realloc#%d(%d)", nallo, n);
		    perror(": ");
		    free(abuf);
		    return((char *)0);
		}
		ebuf = abuf + n;
		cbuf = abuf + o;
	    }
	}
	*cbuf = 0;	/* Terminate buffer */
    }
    else if (ferror(fp)) ;   /* Keep abuf=NULL */
    else abuf = strdup("");

    /* Check error */
    if (ferror(fp)) perror("#***load_stream FAILED");
    return(abuf);
}

char *load_file(char *name, int show)
/*++++++++++++++++
.PURPOSE  Load a file completely into memory
.RETURNS  The loaded text terminated by NULL / NULL if error
.REMARKS  show!=0 to show errors
-----------------*/
{
  struct stat bstat;
  char *buffer;
  FILE *fp;
  int st;

    st = stat(name, &bstat);            /* st==0 ==> File exists...*/
    if ((st!=0) && strncmp(name, "file://", 7) == 0) {
	name += 7;
	st = stat(name, &bstat);
	if (st != 0) name -= 7;
    }
    if ((st == 0) && S_ISREG(bstat.st_mode)) {	/* Regular file... */
      int fd, len, bytes;
        fd = open(name, O_BINARY);
	if (fd<=0) {
	    fprintf(stderr, "#***load_file(%s)", name);
	    perror("");
	    return((char *)0);
	}
	len = lseek(fd, 0, 2); lseek(fd, 0, 0);
	/* Load file into Memory */
	if ((buffer = malloc(len+1))) {
	    bytes = read(fd, buffer, len);
	    if (bytes<len) {
		fprintf(stderr, "#***load_file(%s): read %d/%d\n",
			name, bytes, len);
		if (bytes<0) {
		    perror("#***");
		    free(buffer); buffer = (char *)0;
		}
		else buffer[bytes] = 0;
	    }
	    else buffer[bytes] = 0;
	}
	else {
	    fprintf(stderr, "#***load_file(%s): ", name);
	    perror("");
	}
	close(fd);
	return(buffer);
    }

    /* Not a regular file ==> use load_stream */
    if ((fp = fzopen(name))) {
	buffer = load_stream(fp);
	fzclose(fp);
	return(buffer);
    }

    /* No file found... */
    if (show) { fprintf(stderr, "#***load_file(%s)", name); perror(""); }
    return ((char *)0);
}

char *comfile(char *name)
/*++++++++++++++++
.PURPOSE  Complete a file name with compression extension.
.RETURNS  The complete file name (file exists) / NULL
-----------------*/
{
  static char *buf = (char *)0; static int lbuf = 0;
  struct stat bstat;
  int i, len;

    if (stat(name, &bstat) == 0) 	/* File exists...*/
    	return(name);

    for (i=0; i<ITEMS(zext); i++) {
	len = 2 + strlen(name) + strlen(zext[i]);
	if (len > lbuf) {
	    lbuf = (len|0xf) + 1;
	    if (buf) free(buf);
	    buf = malloc(lbuf);
	}
	strcpy(buf, name); strcat(buf, zext[i]);
	if (stat(buf, &bstat) == 0) 	/* File exists...*/
	    return(buf);
    }
    return((char *)0);
}

char *filename(char *dirname, char *basename, char *ext)
/*++++++++++++++++
.PURPOSE  Generates a filename from its components.
.RETURNS  The full name / NULL
.REMARKS  The name is in an internal buffer
-----------------*/
  /*dirname	(IN) Directory (if unspecified in basename) 	*/
  /*basename	(IN) Name of the file; extension is removed 	*/
  /*ext;	(IN) Extension (the dot should be included);
  				NULL to keep the original extension 	*/
{
  static char *buf = (char *)0; static int lbuf = 0;
  char *p; int i;

  	/* Be sure the buffer to hold the full file name is large enough */
    i = strlen(dirname) + strlen(basename) + (ext ? strlen(ext) : 0) + 3;
    if (i > lbuf) {
	lbuf = (i|15) + 1;	/* Adjust size to multiple of 16 */
	if (buf) free(buf);
	buf = malloc(lbuf);
    }
    	/* Concatenate the three components of the file name */
    strcpy(buf, dirname); p = buf + strlen(buf);
    if (p[-1] != dirsep) *(p++) = dirsep;	/* Insert the /		*/
    if (*basename == dirsep) p = buf;		/* Override if absolute path*/
    strcpy(p, basename);
    	/* If the filename includes an extension, this extension has to be
    	   removed first. Note that the location of the rightmost dot
    	   indicates an extension only if this dot is located AT RIGHT OF
    	   a directory separator (a / on Unix machines)
   	*/
    if (ext) {
	/* Locate the dot of the extension */
	for (i=strlen(basename)-1;
		(i>0) && (basename[i] != '.') && (basename[i] != dirsep);
		i--) ;
    	if ((i>0) && (basename[i] == '.')) p += i;
    	else p += strlen(p);
    	if (*ext != '.') *(p++) = '.';
    	strcpy(p, ext);
    }
    return(buf);
}

static char *command(char **args)
/*++++++++++++++++
.PURPOSE  Create a new string with concatenation (for shell)
.RETURNS  the concatenation
.REMARKS  write the arguments within '
-----------------*/
{
  char *result, *r, **a;
  int len=0;

    for (a=args; *a; a++) len += 3+strlen(*a);
    result = r = malloc(len);
    strcpy(r, args[0]); r += strlen(r);
    for (a=args+1; *a; a++) {
	*(r++) = ' ';
	*(r++) = '\'';
	strcpy(r, *a);  r += strlen(r);
	*(r++) = '\'';
    }
    *r = 0;
    return(result);
}

FILE *fzopen(char *name)
/*++++++++++++++++
.PURPOSE  Open a compressed file for READ ONLY
.RETURNS  The FILE / NULL if error
.REMARKS  Use popen -- Signal SIGCHLD is ignored !
-----------------*/
{
  char *fcat[8];
  FILE *f;
  char *com, *ext;
  int pid, stat, i, ifcat, fd;

    /* Always accepts "-" as stdin */
    if (name && ((name[0] == 0) || ((name[0] == '-') && (name[1] == 0))))
	return(stdin);
    /* V4.6: if extension corresponds to a compression extension,
       remove it
    */
    for (i=strlen(name)-1;
	    (i>0) && (name[i] != '.') && (name[i] != dirsep);
	    i--);
    if ((i>0) && (name[i] == '.')) {
	ext = name+i;
	for (i=0; (i<ITEMS(zext)) && strcmp(ext, zext[i]); i++) ;
	if (i >= ITEMS(zext)) ext = (char *)0;
	else *ext = 0;
    }
    else ext = (char *)0;

    f = fopen(name, RMODE);
    if (ext) *ext = '.';	/* Restore extension */
    if (f) {			/* Standard file     */
	add_info(f, name, 0);
	return(f);
    }

    /* V5.3: accept prefix "file://" */
    if (strncmp(name, "file://", 7) == 0)
	return(fzopen(name+7));

    if ((strncmp(name, "http:", 5) == 0) && (wwwget[0] != 'N')) {
	fcat[0] = wwwget;
	ifcat = 1;
	if (strcmp(wwwget, "wwwget") == 0) {
	    fcat[ifcat++] = "-s";
	    fcat[ifcat++] = "-nostrip";
	    fcat[ifcat++] =  name;
	}
	else if (strcmp(wwwget, "wget") == 0) {
	    fcat[ifcat++] = "-s";
	    fcat[ifcat++] = "-O";
	    fcat[ifcat++] = "-";
	    fcat[ifcat++] =  name;
	}
	else {
	    fprintf(stderr, "#***fzopen(http:...): %s program unknown!\n",
		    wwwget);
	    fprintf(stderr, "#   (report the problem to\t%s\n",
		    "question@simbad.u-strasbg.fr");
	    return(f);
	}
	fcat[ifcat] = (char *)0;
#if 0
	fd = open_pipe(fcat, 0);
	if (fd >= 0) {
	    f = fdopen(fd, RMODE);
	    stat = add_info(f, name, 80);
	    if ((stat<200) || (stat>=300)) { 	/* BAD */
		if ((com = fzinfo(f))) { int c;
		    fprintf(stderr, "#***fzopen(%s)\n", name);
		    for (c='#'; *com; com++) {
		        if (c) fputc(c, stderr);
			fputc(*com, stderr);
			c =  *com == '\n' ? '#' : 0;
		    }
		}
		fzclose(f);
		f = (FILE *)0;
	    }
	    return(f);
	}
#else
	f = popen(com=command(fcat), "r");
	free(com);
	if (f) {
	    stat = add_info(f, name, 80);
	    if ((stat<200) || (stat>=300)) { 	/* BAD */
		if ((com = fzinfo(f))) { int c;
		    fprintf(stderr, "#***fzopen(%s)\n", name);
		    for (c='#'; *com; com++) {
		        if (c) fputc(c, stderr);
			fputc(*com, stderr);
			c =  *com == '\n' ? '#' : 0;
		    }
		}
		fzclose(f);
		f = (FILE *)0;
	    }
	    return(f);
	}
#endif
    }

    if (errno == ENOENT) { 	/* File not found */
#if 0	/* Old code */
    	if (com = comfile(name)) {
	    zcat[zcatfile] = com;
	    fd = open_pipe(zcat, 0);		/* Open for READ */
	    if (fd >= 0) f = fdopen(fd, RMODE);	/* Associate STREAM */
	}
#else
	/* Verify first the file exists !! */
	fcat[0] = "fcat";
	fcat[1] = "-f" ;
    	fcat[2] = name ;
	fcat[3] = (char *)0;
	pid = vfork();
	f = (FILE *)0;				/* File is NULL */
        if (pid < 0) {
	    perror("****fzopen");
	    return(f); 				/* This is NULL	*/
	}
	if (pid) {	/* FATHER */
          GetChildStat:
	    fd = waitpid(pid, &stat, 0) ;
	    if (fd == pid) {
		if (WIFEXITED(stat)) stat = WEXITSTATUS(stat);
		if (stat) return(f); 		/* NULL (failed) */
	    }
	    else if (fd < 0) {			/* Problem ...	 */
		stat = 0 ;			/* Assume OK ..  */
	    }
	    else {				/* Another child... */
		for (i=0; i<ITEMS(children); i++) {
		    if (children[i] == fd) break ;
		}
		if (i<ITEMS(children))
		    chilstat[i] = stat ;
		else fprintf(stderr,
		     "++++fzopen: Got child process #%d, status=0x%X\n",
		     fd, stat) ;
	        goto GetChildStat ;
	    }
	}
	else {
	    execvp(*fcat, fcat);		/* Child just execs */
	    /* Should never go there... */
	    com = getenv("PATH") ;
	    if (!com) com = "(nil)" ;
	    if (getuid() == 0) syslog(LOG_ERR, "***Can't exec %s, PATH=%s\n",
		fcat[0], com) ;
	    else { perror(*fcat); fprintf(stderr, "****PATH=%s\n", com) ; }
	}
	if (stat) return(f) ;			/* Still f=NULL	*/
	fcat[1] = name ;
	fcat[2] = (char *)0 ;	/* Remove the -f option in fcat */
#if 1
	if ((f = popen(com = command(fcat), "r")))
	    add_info(f, name, '%');
	free(com);
#else
	fd = open_pipe(fcat, 0) ;
	if (fd >= 0) {
	    f = fdopen(fd, RMODE);
	    add_info(f, name, '%');		/* zipped */
	}
#endif
#endif
    }
    return (f);
}

char *fzinfo(FILE *file)
/*++++++++++++++++
.PURPOSE  Get the details associated to a file.
.RETURNS  The info (HTTP header, filename...)
.REMARKS  NULL if invalid
-----------------*/
{
  int fd;
    if (!file) return((char *)0);
    fd = fileno(file);
    if (fd<=sizeof(opipes))
	return(header[opipes[fd]]);
    return((char *)0);
}

FILE *fxopen(char *name)
/*++++++++++++++++
.PURPOSE  Open a file for input:
	* it does not exist and starts by '-' or '|' assume shell
	* it does not exist and starts by http:// ==> open url
	* it does not exist and starts by file:// ==> open url
	* it does not exist and starts by &: autostart depending on extension
	* it exists and is executable, execute it.
	* Otherwise open it (eventually via "gawk -f")
.RETURNS  The FILE / NULL if error
.REMARKS  Signal SIGCHLD is ignored !
-----------------*/
{
  struct stat bstat;
  FILE *f=0;
  char *fcat[2];
  int i;

    if (stat(name, &bstat)) {	/* No such file. Shell ? */
	if (name[0] == '<') 	/* V5.3: Added shell */
	    f = fzopen(name+1);
	else if ((name[0] == '-') && (name[1]==0))
	    f = fzopen(name);
	else if ((strchr("|-!^ ", name[0])) &&
		(f = popen(name+1, "r")))
	    add_info(f, name+1, '|');	/* shell opened (via popen) */
	else if (name[0] == '&') { /* V5.4: added file from extension */
            for (i=strlen(name)-1; (i>0) && (name[i] != '.'); i--) ;
	    fcat[0] = name+i;
	    for (i=0; (i<ITEMS(special_ext))
		    && (strcmp(special_ext[i], fcat[0]) == 0); i+=2) ;
	    if (i<ITEMS(special_ext)) {
		fcat[1] = special_ext[i+1];
		fcat[0] = malloc(strlen(fcat[1]) + strlen(name));
		strcpy(fcat[0], fcat[1]); strcat(fcat[0], name+1);
		if ((f = popen(fcat[0], "r")))
		    add_info(f, fcat[0], '|'); /* shell opened (via popen) */
		free(fcat[0]);
	    }
	}
	if (!f) f = fzopen(name);
	return(f);
    }

    /* Verify the file is executable. If yes, execute it. */
    if (bstat.st_mode&0111) {	/* Is executable */
#if 0
        fcat[0] = name; fcat[1] = (char *)0;
        fd = open_pipe(fcat, 0) ;
        if (fd >= 0) {
	    f = fdopen(fd, RMODE);
	    add_info(f, name, '*');	/* Indicates an executable */
        }
        else f = fzopen(name);
#else
	if ((f = popen(name, "r")))
	    add_info(f, name, '|');	/* shell opened (via popen) */
#endif
    }
    if (!f) f = fzopen(name);
    return(f);
}

int fzclose(FILE *file)
/*++++++++++++++++
.PURPOSE  Close a file opened by fzopen
.RETURNS  0 / -1
.REMARKS  Do NOT close stdin (returned status 0)
-----------------*/
{
  int pid, i, fd, status, wait_opt;
    if (file == stdin)
	return(0);
    status = 0 ;
    fd = fileno(file) ;
    if (fd >= sizeof(opipes)) {
	if (fclose(file)) perror("****fzclose") ;
	return(-1);
    }
    i = opipes[fd];
    if (header[i]) {
	status = pclose(file);
	if (status<0) {
	    fprintf(stderr, "#***fzclose(%s)", header[i]+1);
	    perror(":");
	}
	else status = 0;
	free(header[i]);
	header[i] = (char *)0;
    }
    if (chilstat[i] != -1) {	/* Already read... */
        status = chilstat[i] & 0xff ;
        children[i] = 0 ;
        return(status) ;
    }
    wait_opt = WNOHANG;		/* 1st trial for waitpid */
  GetChildStat:
    if (!children[i])          /* No child... */
	return(status);
    kill(children[i], SIGTERM);
    /*fprintf(stderr, "#+++Sent SIGTERM to process#%d\n", children[i]); */
    /* PROBLEM: behind Apache, signal caught by Apache server ?? */
    pid = waitpid(children[i], &status, wait_opt) ;
    if (pid == 0) {  		/* Not Yet changed its state */
	if (wait_opt) fprintf(stderr,
		"#...fzclose's child (pid=%d) not killed\n",
		children[i]);
	wait_opt = 0;
	goto GetChildStat;
    }
    if (pid == children[i]) {
        status &= 0xff ;
        children[i] = 0 ;
        return(status) ;
    }
    if (pid < 0) {		/* Error */
        if (errno == ECHILD) {	/* Still OK */
    	children[i] = 0 ;
    	return(0) ;
        }
        fprintf(stderr, "****fzclose(%d): ", fd), perror("") ;
        return(-1) ;
    }
    else {
        for (i=0; (i<ITEMS(children)) && (children[i] != pid); i++) ;
        if (i < ITEMS(children))
        chilstat[i] = status ;
        goto GetChildStat ;
    }
    return(status) ;
}

int fxclose(FILE *file)
/*++++++++++++++++
.PURPOSE  Close a file opened by fxopen
.RETURNS  0 / -1
-----------------*/
{
    return(fzclose(file));
}

char *fxload(char *name, int show)
/*++++++++++++++++
.PURPOSE  Read the contents of a file, which may be a URL, a shell script,
          an executable file, etc... (see fxopen)
.RETURNS  The contents / NULL if failed
.REMARKS  Signal SIGCHLD is ignored !
-----------------*/
{
  struct stat bstat;
  FILE *fp;

    if ((stat(name, &bstat) == 0) 	/* File exists...*/
      &&(S_ISREG(bstat.st_mode)))	/* Regular file  */
    	return(load_file(name, show));

    /* Not a regular file ==> use load_stream */
    if ((fp = fxopen(name))) { char *buffer;
	buffer = load_stream(fp);
	fzclose(fp);
	return(buffer);
    }

    if (show) fprintf(stderr, "****fxload(%s): no such file\n", name);
    return((char *)0);
}

/*===========================================================================
                Cleanup directories
 *===========================================================================*/
#if DEBUG
#define pr(f,a)	printf(f, a)
#define ign(w)	{ printf(" (%s)\n", w); continue; }
#else
#define pr(f,a)
#define ign(w)	continue
#endif
int fcleaner(char *radix, int age)
/*++++++++++++++++
.PURPOSE  Remove the files starting by radix (in a subprocess) not accessed
		since (age) seconds.
.RETURNS  pid of subprocess
-----------------*/
{
  struct dirent *dire;
  struct stat buf;
  DIR *dirp;
  int len, pid;
  char *dirname, *file_radix;
  time_t timemax;

    /* Start the clean-up in a subprocess */
    pid = fork();
    if (pid < 0) {
        fprintf(stderr, "++++fork(pid=%d) failed!\n", getpid());
        return(-1);
    }
    if (pid) {			/* FATHER */
        signal(SIGCHLD, SIG_IGN);
	/* Store the child PID */
        for (len=0; children[len] && (len<ITEMS(children)); len++ ) ;
	if (len < ITEMS(children)) {
	    children[len] = pid;
	    chilstat[len] = -1 ;
	}
	return(pid);
    }

    dirname = strdup(radix);
    timemax = time(0) - age;
    for (len=strlen(dirname)-1; (len>=0) && (dirname[len] != dirsep); len--) ;
    if (len < 0) 	/* No slash found... */
	file_radix = "";
    else {
	dirname[len++] = 0;
	file_radix = dirname + len;
    }

#if DEBUG
    printf("#...moving to directory: %s\n#   (age=%d, file_radix=%s) U%d\n",
	    dirname, age, file_radix, timemax);
#endif
    if (chdir(dirname)) {
	fprintf(stderr, "****Can't chdir(%s)", dirname);
	perror("");
	exit(1);
    }

    /* Scan the directory */
    dirp = opendir(".");
    if (!dirp) {
	fprintf(stderr, "****cleanup(%s/%s*)", dirname, file_radix);
	perror("");
	exit(1);
    }

    /* Remove files having the time of last access too old */
    len = strlen(file_radix);
    while ((dire = readdir(dirp))) {
	pr("#...file %s:", dire->d_name);
	if (strncmp(dire->d_name, file_radix, len) != 0)
	    ign("mismatching name");
	if (lstat(dire->d_name, &buf) < 0)	/* Bad entry ??? */
	    ign("***lstat failed***");
	pr(" atime=%d", buf.st_atime);
	if (!S_ISREG(buf.st_mode))		/* Not a regular file */
	    ign("not a regular file");
	if (buf.st_atime >= timemax)		/* recently accessed  */
	    ign("too young");
	pr(" %s\n", "REMOVE");
	unlink(dire->d_name);
    }
    closedir(dirp);
    exit(0);
}

#ifdef TEST
static char inbyte;

int main()
/*++++++++++++++++
.PURPOSE  Just a TEST program
-----------------*/
{
  static char buf[256]; FILE *f;
  struct stat bstat;
  int i, s; int n;
  char *b, *p, *radix;

    for (i=0; i<3; i++) {
    	if (fstat(i, &bstat)) perror("????");
	printf("File %d: mode=%o, size=%d\n",
		i, bstat.st_mode, (int)bstat.st_size);
    }

    /* Tests cleaner */
    printf("Test fcleaner: which radix for files to clean? ");
    strcpy(buf, "ls -l "); s = strlen(buf);
    if (!fgets(buf+s, sizeof(buf)-s-2, stdin)) exit(0);
    for (i=strlen(buf)-1; (i>0) && isspace(buf[i]); i--)
	buf[i] = 0;
    radix = strdup(buf+s);
    strcat(buf, "*");
    if ((f = popen(buf, "r"))) {
	while (fgets(buf, sizeof(buf), f))
	    printf("%s", buf);
	pclose(f);
    }
    else perror("#***");
    printf("#...Launch cleaner for files older than (secs) ");
    if (!fgets(buf, sizeof(buf), stdin)) exit(0);
    s = fcleaner(radix, atoi(buf));
    printf("#...pid(cleaner)=%d\n", s);

    while(1) {
    	printf("File to open: ");
	if (!fgets(buf, sizeof(buf), stdin)) break;
	for (i=strlen(buf)-1; (i>0) && isspace(buf[i]); i--)
	    buf[i] = 0;
	if (is_dir(buf)) {
	    printf("****%s is a directory !\n", buf);
	    continue;
	}
    	f = fzopen(buf);
	if (! f) {
	    fprintf(stderr, "****Can't open file ");
	    fprintq(stderr, buf);
	    fputc('\n', stderr);
	    continue;
	}
	if ((p = fzinfo(f)))
	     fprintf(stderr, "#===fzinfo of file is:\n%s\n", p);
	else fprintf(stderr, "++++No info available!\n");
    	if (fstat(fileno(f), &bstat)) perror("????");
    	if (! f) { perror(buf); continue; }
	printf("File %d: mode=%o, size=%d\n", fileno(f),
		bstat.st_mode, (int)bstat.st_size);
    	i = 3;
    	printf("List of the first %d buffers of file %s:\n", i, buf);
    	while (--i >= 0) {
    	    n = fread(buf, 1, sizeof(buf), f);
    	    if (n < 0)  { perror("Problem"); break; }
    	    if (n == 0) { printf("====EOF====\n"); break; }
    	    printf("Read %d bytes:", n);
    	    inbyte = getchar();
    	    b = p = buf;
    	    while (--n >= 0) {
    	    	printf("%4o", *p & 0xff);
    	    	if (*p == '\n') {
    	    	    s = (p-b)+1;
    	    	    printf(" (%d bytes)\n", s);
    	    	    s = fwrite(b, 1, s, stdout);
    	    	    fflush(stdout);
    	    	    b = p+1;
    	    	}
    	    	p++;
    	    }
    	    printf("\n"), fwrite(b, 1, p-b, stdout), printf("\n");
    	}
    	if (fstat(1, &bstat)) perror("????");
	printf("File 1: mode=%o, size=%d\n",
		bstat.st_mode, (int)bstat.st_size);
	fclose(f);
    }
    wait_pipe(0);
    exit(0);
}
#endif
