/*++++++++++++++
.IDENTIFICATION tabutil.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   15-May-1991 (anafile)
.VERSION  4.0   28-Jul-1994 New package
.VERSION  4.1   15-Nov-1994: Verify findformat
.VERSION  4.2   01-Jun-1995: Section "Related Data"
.VERSION  4.3   11-Jun-1997: findfile: forget about the extensions...
.VERSION  4.4   03-Nov-2002: findlabel: accept a FILINFO structure
.VERSION  4.5   17-Feb-2004: findfile: accept subdirectory in filename
.VERSION  4.6   12-Apr-2004: cad_paste (horizontal merge)
.VERSION  4.7   03-Aug-2004: sec_findnote: accept a note with *
.VERSION  4.8   28-Jan-2005: tbf_findcol, accept column "col@table"
.VERSION  4.9   27-May-2005: cad_paste with a different number of records is
			a WARNING only
.VERSION  4.91  11-Feb-2006: Join columns where one is ignored (F_NODB)
.VERSION  4.92  06-May-2006: Bug in cad_paste fixed (move of position columns)
.VERSION  4.93  16-Jul-2006: Bug in cad_paste fixed (nj count)
.VERSION  4.94  11-Jan-2007: Generic filenames in cad_findfile
.VERSION  4.95  02-Jun-2007: Cosmetics (-Wall)
.VERSION  4.96  10-Sep-2007: Bug in cad_paste fixed (nj count)
.VERSION  4.97  14-Sep-2007: Another bug in cad_paste (equinox/epoch)
.VERSION  4.98  20-Jan-2008: cad_paste: bug in case of ignored columns
.VERSION  4.99  30-Jan-2008: cad_paste: another buf in case of ignored columns
.VERSION  5.0   05-Mar-2008: cad_paste: remove MPOS flag if several coordinates
.VERSION  5.01  14-Oct-2009: SEC_History
.VERSION  5.02  28-Oct-2009: findlkabel in case of file without table
.VERSION  5.03  15-Dec-2009: 64-bit
.VERSION  5.04  02-Sep-2010: cad_paste revisited
.VERSION  5.05  05-Jan-2011: cad_paste fixed.
.VERSION  5.06  13-Jan-2011: tbf_findlabel accepts +offset / #(col#)
.VERSION  5.07  17-Jul-2011: cad_findfile set F_INDX for generic file
.VERSION  5.08  02-Aug-2011: cad_paste crash if file not described
.VERSION  5.09  01-Jan-2012: accept .gz in cad_findfile
.VERSION  5.10  11-Mar-2013: fmt_edrange of null values
.COMMENTS       Some utilities with TABxxx structures
---------------*/

#include <catstd.h>

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

#ifdef DEBUG
static char inbyte;
#endif

#ifndef NULL8
#define NULL8 0x8000000000000000LL
#endif
#ifndef int8l
#define int8l 2
#endif

#define MAX_PREFIX 14
static char secid[MAX_PREFIX+22] = "sRM" ;	/* Section of ReadMe */
static int lprefix = 3 ;

static char *atype[] = {        /* File Type     */
	"ascii bulk", "ascii file", "ascii mailable", "ascii binary",
	"ebcdic", "fits", "hdf", "binary", "directory",
};

/*==================================================================
		Editions
 *==================================================================*/
char *ftype(btyp)
/*++++++++++++++++
.PURPOSE  Interpret the File Type Symbol
.RETURNS  The ascii equivalent
-----------------*/
  int btyp;	/* IN: File type Symbol */
{
    if (btyp == ITEMS(atype)) return("\\n");
    if (btyp <0 || btyp >= ITEMS(atype))
    	return("???");
    return(atype[btyp]);
}

char *strfmt(text)
/*++++++++++++++++
.PURPOSE  Interpret the text, derive a fromat (e.g. F12.6)
.RETURNS  The Fortran Format
.REMARKS  in a static buffer
-----------------*/
  char *text;	/* IN: The text to examine */
{
  static char ffmt[20];
  int len, dec;
  char *p, *d, f;
    len = strlen(text);
    sprintf(ffmt, "A%d", len);
    while (*text == ' ') text++;
    p = text;
    f = 'I'; dec=0;
    if ((*p == '-') || (*p == '+')) p++;
    while (isdigit(*p)) p++;
    if (*p == '.') {
	f = 'F';
	d = ++p;
	while (isdigit(*p)) p++;
	if (tolower(*p) == 'e') {
	    f = 'E';
	    ++dec;	/* Number of significant digits */
	    ++p;
	    if ((*p == '-') || (*p == '+')) p++;
	    while (isdigit(*p)) p++;
	}
	dec += p-d;
	len = p-text;
    }
    while (*p == ' ') p++;
    if (*p) 	/* Can't be a number */
        return(ffmt);
    ffmt[0] = f;
    sprintf(ffmt+1, "%d", len);
    if (dec>0) sprintf(ffmt+strlen(ffmt), ".%d", dec);
    return(ffmt);
}

char *seclass(class)
/*++++++++++++++++
.PURPOSE  Interpret the section class (SEC_xxx)
.RETURNS  Its value
-----------------*/
  int class;	/* IN: Section Class type Symbol */
{
  char *p;
    switch(class) {
    case SEC_Title:	  p="Title"; 		break;
    case SEC_Description: p="Description"; 	break;
    case SEC_Abstract:    p="Abstract";		break;
    case SEC_References:  p="References"; 	break;
    case SEC_History:     p="History"; 		break;
    case SEC_FileSummary: p="File Summary"; 	break;
    case SEC_BytebyByte:  p="Byte-by-byte"; 	break;
    case SEC_Note:	  p="Note"; 		break;
    case SEC_Objects:	  p="Objects";		break;
    case SEC_Keywords:	  p="Keywords"; 	break;
    case SEC_SeeAlso:	  p="Related Data"; 	break;
    case SEC_Naming:	  p="Nomenclature";	break;
    case SEC_End:	  p="(End)"; 		break;
    case SEC_Col:	  p="CSV Description";	break;
    case SEC_Unknown:	  p="(other)"; 		break;
    case SEC_Ack:	  p="Acknowledgements"; break;
    default:		  p=(char *)0; 		break;
    }
    return(p);
}

/*==================================================================*/
char *fmt_ed(format)
/*++++++++++++++++
.PURPOSE  Edit the Fortran Format
.RETURNS  Edited format WITHOUT the repetition factor
-----------------*/
  COLINFO *format;	/* IN: Fortran Format */
{
  static char buf[16];
  char *p;

    if (!format) return("");
    if (format->len == 0) return("");
    p = buf;
    if (format->rep > 1) sprintf(p, "%d", format->rep), p += strlen(p);
    *(p++) = format->ff;
    if (format->flags&F_ZERO) *(p++) = '0' ;
    sprintf(p, "%d", format->len);
    switch(format->ff) {
      case 'F': case 'E': case 'D':
    	p += strlen(p);
	sprintf(p, ".%d", format->len - format->odp -1);
	break;
    }
    return(buf);
}

char *fmt_edrange(format)
/*++++++++++++++++
.PURPOSE  Edit the Range
.RETURNS  Edited range
-----------------*/
  COLINFO *format;	/* IN: Fortran Format */
{
  static char buf[256]; int i;
  char *p, kp, km;
  double *dlim;
  int8   *llim;

    if (!format) return("");
    p = buf;
    switch(format->ff) {
      case 'A':
        km = format->range['-']; kp = format->range[']'];
        format->range['-'] = 0;  format->range[']'] = 0;
        if (kp) *(p++) = ']';
        for (i=1; i<256; i++) {
            if(!format->range[i]) continue;
#if O_BINARY
            if(i == '\r') continue;
#endif
            *(p++) = i;
            if (format->range[i+1] && format->range[i+2]) {
            	while(i<256 && format->range[i]) i++;
            	*(p++) = '-'; *(p++) = --i;
            }
        }
        if (km) *(p++) = '-';
        format->range['-'] = km; format->range[']'] = kp;
        *p = 0;
        break;
      case 'I':
        llim = (int8 *)format->range;
	if (llim[1] == NULL8) return("");  
#if int8l == 2
        sprintf(buf, "%lld/%lld", llim[0], llim[1]);
#else
        sprintf(buf, "%ld/%ld", llim[0], llim[1]);
#endif
        break;
      default:
        dlim = (double *)format->range;
        sprintf(buf, "%g/%g", dlim[0], dlim[1]);
        break;
    }
    return(buf);
}

/*==================================================================*/

int sec_prhead(ps, fout)
/*++++++++++++++++
.PURPOSE  Print out a section header
.RETURNS  Number of lines
-----------------*/
  SECTION *ps;	/* IN: Section to be printed */
  FILE   *fout;	/* IN: Opened output file */
{
  int lines, bytes;
  char *p;

    lines = 1;
    p = seclass(ps->class);
    fprintf(fout, "----Line%4d: '%s' section ", ps->lineno+1, p);
    bytes = 25+strlen(p);
    if (ps->used & SECNOTE_used) bytes += 7, fprintf(fout, "(used) ");
    fprintf(fout, "made of %d bytes after ", (int)(ps->ebody - ps->abody));
    bytes += 24;
    if (ps->header) {
    	if ((bytes+(int)strlen(ps->header)) > 80)
    	     lines++, fprintf(fout, "\n\t");
    	fprintf(fout, "\"%s\"", ps->header);
    }
    fprintf(fout, "\n");
    return(lines);
}

/*==================================================================
		Lookup for various items from names
 *==================================================================*/

void sec_prefix(text)
/*++++++++++++++++
.PURPOSE  Set the prefix for section edition
.RETURNS  The previously defined prefix.
.REMARKS  A local copy of text is taken. Text truncated if too long.
-----------------*/
  char *text ;	/* IN : new prefix */
{
    lprefix = strlen(text) ;
    if (lprefix > MAX_PREFIX) lprefix = MAX_PREFIX ;
    strncpy(secid, text, lprefix) ; secid[lprefix] = 0 ;
}

char *sec_id(obj)
/*++++++++++++++++
.PURPOSE  Edit the Section ID (for \\label / \\ref)
.RETURNS  The edited sRM#.#  (in a static buffer)
-----------------*/
  void *obj;	/* Section to ID */
{
  static char no_signature[] = "\3" ;
  SECTION *sec ;
  char *signature ;
  intp i ;

    if (!obj)	/* Get Previous definition */
	return(secid) ;

    sec = (SECTION *)obj ;
    signature = sec->signature ;
    if (!signature) signature = no_signature ;
    switch (signature[1]) {
      case 'S': 		/* It's a section */
        sprintf(secid+lprefix, "%d.%d", sec->class, sec->secno) ;
	return(secid) ;
      case 'T':			/* A table, take its section */
	if ((sec = ((TABINFO *)obj)->section))
	    i = sec->secno ;
	else i = (intp) obj ;
	break ;
      case 'F':			/* A file, take its lineno   */
	i = ((FILINFO *)obj)->lineno ;
	break ;
      case 'C':			/* A column, take its lineno */
	i = ((COLINFO *)obj)->lineno ;
	break ;
      default:
	i = (intp) obj ;
    }
    sprintf(secid+lprefix, "-%c.%d", signature[1], (int)i) ;
    return(secid) ;
}

SECTION *sec_find(sec, text)
/*++++++++++++++++
.PURPOSE  Find a Section which starts by text
.RETURNS  Section / NULL
-----------------*/
  SECTION *sec;	/* IN: First Section to look at */
  char *text;	/* IN: Starting text of title	*/
{
  SECTION *ps;
  int len ; char *p ;

    len = strlen(text) ;
    for (ps = sec; ps ; ps = ps->next) {
	p = ps->header ;
	if (!p) continue ;
	if (strncmp(p, text, len) == 0) return(ps) ;
    }
    return((SECTION *)0) ;
}

SECTION *sec_findnote(sec, name)
/*++++++++++++++++
.PURPOSE  Find a "Note" section which header contains "name"
.RETURNS  Section / NULL
-----------------*/
  SECTION *sec;	/* Section to look at */
  char *name;	/* IN: Name to look at */
{
  SECTION *ps;
  int local_note, ok;
  char x, *p, *e;

    if ((name[0] == '(') && (!isdigit(name[1]))) 	/* GLOBAL Note	   */
	local_note = 0 ;
    else local_note = 1 ;
    ps = sec;
    if (ps->class != SEC_Note) ps = ps->next;

    while(ps) {
    	if (local_note && (ps->class != SEC_Note)) break;
	e = ps->header + strloc(ps->header, ':');	/* Keep first part */
	x = *e, *e = 0;					/* Save separator  */
	p = ps->header + strloc(ps->header, ' ');	/* Skip 'Note'     */
	while (isspace(*p)) p++;
	if (*p != '(') { while (isgraph(*p)) p++; while (isspace(*p)) p++; }
	ok = match_word(p, name, 0);
	if ((!ok) && strchr(p, '*')) 		/* V4.7:  Accept a *-match */
	    ok = strmatch(p, name);
	*e = x;						/* Restore separat */
	if (ok) return(ps);
    	ps = ps->next;
    }
#ifdef DEBUG
    	printf("....Note '%s' not found...", name);
    	inbyte = getchar();
#endif
    return((SECTION *)0);
}

int tbf_text(void *obj, char *text)
/*++++++++++++++++
.PURPOSE  Change the text of a colum/table/catalog explanation
.RETURNS  0=OK / -1=error
-----------------*/
{
  COLINFO *col;
    if (!obj) return(-1);
    col = (COLINFO *)obj;
    if (col->signature == cad_signature("\3COL")) {	/* Column */
	if (col->flags&F_Atext) free(col->text);
	col->text = strdup(text);
	return(0);
    }
    if (col->signature == cad_signature("\3FIL")) {  	/* File */
 	FILINFO *fil = (FILINFO *)obj;
	if (fil->flags&F_Atext) free(fil->text);
	fil->text = strdup(text);
	return(0);
    }
    fprintf(stderr, "#+++tbf_text(%p, %s) apllies only on a column or file\n",
	    obj, text);
    return(-1);
}

COLINFO *tbf_findcol(obj, name)
/*++++++++++++++++
.PURPOSE  Find a label matching a template
.RETURNS  Format / NULL
.REMARKS  Name may be  col@table
-----------------*/
  void  *obj;	/* IN: a CATINFO : FILINFO : TABINFO */
  char *name;	/* IN: Label to look at */
{
  CATINFO *cat=(CATINFO *)0;
  TABINFO *tab=(TABINFO *)0;
  COLINFO *col=(COLINFO *)0;
  char *pat, *fname;
  int len=0, i;

    if (!obj) return((COLINFO *)0);
    tab = (TABINFO *)obj;
    if ((pat = strchr(name, '@'))) {
	len = (pat - name);		/* Length of column name */
	pat++;				/* Name of table */
    }
    /* Find a column in the whole catalog */
    if (tab->signature == cad_signature("\3CAT")) {
	cat = (CATINFO *)obj;
	if (pat) {			/* Locate the file */
	    name[len] = 0;
	    col = tbf_findlabel((TABINFO *)cad_findfile(cat, pat), name);
	    name[len] = '@';
	}
	else for (tab=cat->atinfo, i=cat->ntinfo; --i>=0; tab++) {
	    if ((col = tbf_findlabel(tab, name))) break;
	}
	return(col);
    }
    if (tab->signature == cad_signature("\3FIL")) {
	if (pat) { 	/* Verify the correct file name */
	    fname = ((FILINFO *)obj)->dfile;
	    while (*fname == *pat) fname++, pat++;
	    if (*pat) return((COLINFO *)0);
	    if (*fname && (*fname == '.')) return((COLINFO *)0);
	    name[len] = 0;
	}
	col = tbf_findlabel(tab, name);
	if (pat) name[len] = '@';
	return(col);
    }
    if (tab->signature == cad_signature("\3TAB")) {
	col = tbf_findlabel(tab, name);
	return(col);
    }
    return((COLINFO *)0);
}

COLINFO *tbf_findlabel(T, name)
/*++++++++++++++++
.PURPOSE  Find a label matching a template
.RETURNS  Format / NULL
.REMARKS  When name starts by
	* $ ==>> just match column with name starting by the name
	* + ==>> column starting at byte offset specified
	* # ==>> column #x (in the description)
-----------------*/
  TABINFO *T;	/* IN: "Byte-by-Byte description" */
  char *name;	/* IN: Label to look at */
{
  COLINFO *pf, *pe; char *p, *n; int icol;

    if (!T) return((COLINFO *)0);
    if (T->signature[1] == 'F')		/* I've a FILINFO structure */
	T = ((FILINFO *)T)->tabform;

    if (!T) /* File without table description, V5.02 */
	return((COLINFO *)0);

    pf = T->afmt ; pe = pf + T->nfmt ;
    switch(*name) {
      case '$': 	/* Keep what comes first */
	name++ ;
	while (pf < pe) {
	    for (n=name, p=pf->label; *n && (*n == *p) ; n++, p++ ) ;
	    if (! *n) return(pf) ;
	    pf++ ;
	}
	return((COLINFO *)0);
      case '#':		/* Column number */
	icol = atoi(name+1);
	pf =  (icol>0) && (icol<=T->nfmt) ? pf + (icol-1) : 0;
	return(pf);
      case '+':		/* Column offset */
	icol = atoi(name+1);
	while (pf < pe) {
	    if (pf->b0 == icol) return(pf);
	    pf++;
	}
	return((COLINFO *)0);
    }

    while (pf < pe) {
	if (strdiff(pf->label, name) == 0) return(pf);
	pf++ ;
    }

    return((COLINFO *)0);
}

FILINFO *cad_findfile(T, filename)
/*++++++++++++++++
.PURPOSE  Find the description of a file
.RETURNS  FILINFO corresponding to filename;
	#NO# a new one is allocated if necessary.
.REMARKS  "ReadMe" file name or unspecified returns NULL
	The ".dat" extension is not required.
	F_INDX flag set if matched from template (generic file)
-----------------*/
  CATINFO *T;		/* IN: FormatFile description  	*/
  char *filename;	/* IN: Name of data file	*/
{
  FILINFO *pf; int i, generic;
  char *p, *f ;

#ifdef DEBUG
    printf("....findfile in %p, afinfo=%p, nfinfo=%d\n", T,
    	T->afinfo, T->nfinfo);
#endif
    if (!filename) return((FILINFO *)0);
    for (pf = T->afinfo, i = T->nfinfo; --i >= 0; pf++) {
	generic = 0;		/* V5.07: set if generic match */
	for (p = pf->dfile, f = filename; *p && (*p == *f); p++, f++) ;
	if (*p == '*') {	/* V4.5, mod. 4.94: Accept generic */
	    generic = F_INDX;
	    if (strmatch(p, f))
	        p += strlen(p);
	    else p++;
	    f += strlen(f);
	}
	pf->flags |= generic;	/* V5.07: flag set (generic) */
	if (*f == 0) {
	    if (*p == 0) return(pf);
	    if (*p == '.') 	/* Ignore the extension ! */
	    	return(pf);
	}
	if ((*p == 0) && (*f == '.')) { /* Added V5.09: accept compressed */
	    if (strchr(f, 'z')) return(pf);
	}
	/*if (strdiff(pf->dfile, filename) == 0) return(pf);*/
    }
    return((FILINFO *)0);
}

FILINFO *cad_addfile(T, filename)
/*++++++++++++++++
.PURPOSE  Add a new file description
.RETURNS  FILINFO corresponding to existing filename;
	a new one is allocated if necessary.
.REMARKS  "ReadMe" file name or unspecified returns NULL
-----------------*/
  CATINFO *T;		/* IN: FormatFile description  	*/
  char *filename;	/* IN: Name of data file	*/
{
  FILINFO *pf;

#ifdef DEBUG
    printf("....findfile in %p, afinfo=%p, nfinfo=%d\n", T,
    	T->afinfo, T->nfinfo);
#endif
    if ((pf = cad_findfile(T, filename))) return(pf);

    /* Don't add the "ReadMe" file */
    if (T->ffile) {
    	if (strdiff(T->ffile, filename) == 0) return((FILINFO *)0);
    }

    /* Find a slot for the additional file */
    if (T->mfinfo == T->nfinfo) {
    	T->mfinfo += 16;
    	T->afinfo = (FILINFO *)realloc(T->afinfo,
    	   T->mfinfo*sizeof(FILINFO));
	if (!T->afinfo) fprintf(stderr,
	   "****Can't realloc memory for %d files\n", T->mfinfo),
	   terminate((char *)0);
	memset(T->afinfo+T->nfinfo, 0, (T->mfinfo-T->nfinfo)*sizeof(FILINFO));
    }

    /* Initialize (non-zero items) */
    pf = T->afinfo + T->nfinfo; T->nfinfo += 1;
    pf->signature = cad_signature("\3FIL");
    pf->oid   = cad_oid();
    pf->dfile = filename;
    pf->reclen = pf->records = pf->flags = pf->lineno = 0;
    pf->note = (SECTION *)0;
    pf->tabform = cad_findformat(T, filename);
    pf->equinox = pf->epoch = (char *)0;
    pf->text = pf->more = pf->comment = (char *)0;

    return(pf);
}

FILINFO *cad_paste(T, list)
/*++++++++++++++++
.PURPOSE  Create a new file, pasted schema of files mentioned in list
.RETURNS  FILINFO corresponding to a new file, with the union of fields
.REMARKS  List of files in the set are separated by blank or comma.
	The new file, as well as each in the list, is flagged F_CATA,
	and the common columns are flagged by F_JOIN
-----------------*/
  CATINFO *T;		/* IN: FormatFile description  	*/
  char *list;		/* IN: The list of pasted files	*/
{
  FILINFO *pf, *new;
  TABINFO *pt, *tab;
  COLINFO *pc, *ec, *rc, tcol;
  short posfmt[8];
  char buf[1024], bufmt[32];
  char *p, *b, *e, x;
  int j, n, c, ipos, iname, nj;

#ifdef DEBUG
    printf("....cad_paste into %s: %s\n", filename, list);
#endif

    /* Create the new name: use list of files, separate by a comma */
    b = buf; e = buf+sizeof(buf)-2;
    p=list;
    while (*p && (b<e)) {
	if ((*p == ',') || isspace(*p)) {
	    *(b++) = ',';
	    while (isspace(*p)) p++;
	    continue;
	}
	*(b++) = *(p++);
    }
    *b = 0;
    if (b>=e) {
	fprintf(stderr, "****cad_paste(%s) truncated to %s\n", list, buf);
	return ((FILINFO *)0);
    }
    if ((b>buf) && (b[-1]==',')) *--b = 0;

    /* Verify all files exist, with identical record counts,
       and count upper bound of how many columns with be in the result
    */
    for (b=buf, n=0, c=0; *b; b += iname) {
	iname = strloc(b, ',');
	if (iname==0) { iname=1; continue; }
	x = b[iname]; b[iname] = 0;
	pf = cad_findfile(T, b);
	if (!pf) {
	    fprintf(stderr, "****cad_paste(%s): couldn't locate '%s'\n",
	      list, b);
	    b[iname] = x;
	    return((FILINFO *)0);
	}
	b[iname] = x;
	if (n == 0) n = pf->records;
	else if (n != pf->records) fprintf(stderr,
	    "++++cad_paste(%s): %s(%d) #records!=%d!\n",
	    buf, pf->dfile, pf->records, n);
	if ((pt = pf->tabform)) c += pt->nfmt;
    }

    /* Create the new file + Table */
    new = cad_addfile(T, strdup(buf));
    new->records = n;
    new->flags |= F_CATA;	/* File created by paste */
    new->tabform = tab = malloc(sizeof(TABINFO));
    new->flags |= F_Aunit;	/* tabform was allocated */
    memset(tab, 0, sizeof(TABINFO));
    tab->signature = cad_signature("\3TAB");
    tab->oid       = cad_oid();
    memset(tab->posfmt, -1, sizeof(tab->posfmt));
    tab->afmt = (COLINFO *)malloc(c*sizeof(COLINFO));

    /* Insert all fields. Here n is the lrecl offset */
    for (b=new->dfile, n=0, c=0, nj=0; *b; b += iname) {
	iname = strloc(b, ',');
	if (iname==0) { iname=1; continue; }
	x = b[iname]; b[iname] = 0;
	pf = cad_findfile(T, b);
	pt = pf->tabform;
	if (!pt) {	/* File without a description? (V5.08) */
	    fprintf(stderr, "#***cad_paste: file '%s' has no description!\n",
		    pf->dfile);
	    tbf_free(tab);
	    return((FILINFO *)0);
	}
	pf->flags |= F_CATA;		/* Mark the file */
	b[iname] = x;
	/* Add missing items */
	if (!new->note) new->note = pf->note;
	if (!new->text) new->text = pf->text;
	if (!new->more) new->more = pf->more;
	if (!new->comment) new->comment = pf->comment;
	/* Install Equinox + Epoch */
	if ((tab->posfmt[0] == -1) && (pt->posfmt[0]>=0)) {
	    pc = pt->afmt + pt->posfmt[0];
	    if (pc->flags&F_NODB) 	/* A main position to be ignored??? */
		fprintf(stderr,
		    "#+++cad_paste(%s) has MAIN_POS marked NODB: %s\n",
			pf->dfile, pc->label);
	    new->epoch   = pf->epoch;
	    new->equinox = pf->equinox;
	    if (new->epoch)   new->epoch   = strdup(new->epoch);   /* V4.97 */
	    if (new->equinox) new->equinox = strdup(new->equinox); /* V4.97 */
	    memcpy(posfmt, pt->posfmt, sizeof(tab->posfmt));
	}
	else posfmt[0] = -1;		/* No position to transfer */
	ipos = 0;
	for (pc=pt->afmt, ec=pc+pt->nfmt; pc<ec; pc++) {
	    c = tab->nfmt;		/* New column number */
	    /* Take care of adding the details on the coordinates in table */
	    if ((posfmt[0] >= 0) && (pc->flags&F_MPOS)) {    /* Mod. V4.99 */
		j = pc-pt->afmt;
		for (ipos=0; (posfmt[ipos]>=0) && (posfmt[ipos]!=j); ipos++) ;
		if (posfmt[ipos] < 0) fprintf(stderr,
		    "++++cad_paste(%s) column marked F_MPOS ignored: %s\n",
		    pf->dfile, pc->label);
		else tab->posfmt[ipos] = c;
	    }
	    tab->afmt[c] = *pc;		/* Copy column def */
	    tab->afmt[c].b0 += n;
	    tab->afmt[c].b1 += n;
	    tab->afmt[c].flags &= ~(F_Atext|F_Aname|F_Aunit);
	    tab->afmt[c].flags |= F_refer; /* lims, value etc not allocated */
	    if ((posfmt[0] < 0) && (pc->flags&F_MPOS)) {
		/* V5.00: remove MPOS flag when position is not main */
		fprintf(stderr,
			"++++cad_paste(%s), column %s: remove F_MPOS\n",
			pf->dfile, pc->label);
		tab->afmt[c].flags &= ~F_MPOS;
	    }
	    /* Mark the JOIN fields, and verify compatibility
	       BUT columns "--" are not required to be identical !
	    */
	    if (pc->label[0] == '-') rc = (COLINFO *)0;
	    else rc = tbf_findlabel(tab, pc->label);
	    if (rc) {				/* 2 columns with same name */
		/* Old column having the same name may be NODB:
		 * in this case, push this old column AFTER this one.
		 */
		if ((rc->flags&F_NODB) && ((tab->afmt[c].flags&F_NODB) == 0)) {
		    ipos = (rc - tab->afmt);	/* V5.5: beginning of shift */
		    memcpy(&tcol, rc, sizeof(COLINFO));
		    while (rc < &(tab->afmt[c])) {
			memcpy(rc, rc+1, sizeof(COLINFO));
			rc++;
		    }
		    memcpy(&(tab->afmt[c]), &tcol, sizeof(COLINFO));
		    /* V5.5: Reflect the shift in the list of MPOS components */
		    for (j=0; (j<ITEMS(tab->posfmt)) && (tab->posfmt[j]>=0);
			 j++) {
			if (tab->posfmt[j] > ipos) tab->posfmt[j] -= 1;
		    }
		}
		tab->afmt[c].flags |= F_JOIN;
		strncpy(bufmt, fmt_ed(rc), sizeof(bufmt));
		/* Verify editing formats are compatible */
		p = fmt_ed(&tab->afmt[c]);
		if ((tab->afmt[c].flags&F_NODB)|(rc->flags&F_NODB)) ;
		else if (strcmp(bufmt, p) != 0) {
		    fprintf(stderr,
		     "****cad_paste(%s), column '%s' has formats '%s' / '%s'\n",
		     list, rc->label, bufmt, p);
		    tbf_free(tab);
		    T->nfinfo -= 1;	/* ignore the file */
		    return((FILINFO *)0);
		}
	    }
	    /* Counts of JOIN columns: in case of a join of joins,
	       the old column used in a join is still used in a join...
	    */
	    if (tab->afmt[c].flags&F_JOIN)
		nj++;			/* Counter of joins   */
	    /* fprintf(stderr, "#...tab->afmt[0]=%s\n", tab->afmt[0].label); */
	    tab->nfmt += 1;
	}
	n += pf->reclen;
    }
    new->reclen = n;			/* Merged record length */

    /* Write the JOIN (and NODB) fields at the end */
    for (pc = tab->afmt + tab->nfmt -1; --nj>=0; pc--) {
        if (pc->flags&F_JOIN) continue;
        /* Find the join, and move it to the end */
        for (rc=pc; (rc->flags&F_JOIN) == 0; --rc) ;
	if (rc<tab->afmt) {
	    fprintf(stderr, "\n****cad_paste(%s): fatal error join#%d (%s)\n",
		    new->dfile, nj, pc->label);
	    return((FILINFO *)0);
	}
	/* fprintf(stderr, "#...moving join #%d: %s\n", nj, rc->label); */
        memcpy(&tcol, rc, sizeof(COLINFO));		/* Save join col. */
        while (rc < pc) {
    	    memcpy(rc, rc+1, sizeof(COLINFO));
    	    rc++;
	    /* Verify a possible move of a position column... */
	    if (rc->flags&F_MPOS) {
	        j = rc-tab->afmt;
	        for (ipos=0; (ipos<ITEMS(posfmt))
			&& (tab->posfmt[ipos]>=0) && (tab->posfmt[ipos]!=j);
		    ipos++) ;
		if (ipos < ITEMS(posfmt)) tab->posfmt[ipos] -= 1;
		else fprintf(stderr,
		    "****cad_paste(%s): Pb with position '%s'\n",
		    list, rc->label);
	    }
        }
        memcpy(pc, &tcol, sizeof(COLINFO));
    }

    /* Remove the 'join' columns, but add them to the 'lineno' */
    for (pc=tab->afmt+tab->nfmt-1; (pc>tab->afmt) && (pc->flags&F_JOIN); pc--) {
	tab->lineno -= 1;
	tab->nfmt -= 1;
    }

    return(new);
}


TABINFO *cad_findformat(T, filename)
/*++++++++++++++++
.PURPOSE  Find a label matching a template
.RETURNS  Format / NULL
-----------------*/
  CATINFO *T;		/* IN: FormatFile description  	*/
  char *filename;	/* IN: Name of data file	*/
{
  TABINFO *pt; SECTION *ps; FILINFO *pf; int i;

#ifdef DEBUG
    printf("....findformat in %p, atinfo=%p, ntinfo=%d, file=%s\n", T,
    	T->atinfo, T->ntinfo, filename);
#endif
    for (pt = T->atinfo, i = T->ntinfo; --i >= 0; pt++) {
	ps = pt->section;
	if (!ps->header) return(pt);
	if (match_word(ps->header, filename, ':')) return(pt);
    }
    if ((pf = cad_findfile(T, filename))) return(pf->tabform);
    return((TABINFO *)0);
}
