/*++++++++++++++ 
.IDENTIFICATION tofits.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    Unix, CDS Catalogues
.KEYWORDS       Table, FITS
.VERSION  1.0   15-Jul-1993: C-shell version(mods 1.1 to 1.9 deleted)
.VERSION  2.0   02-Aug-1994: C package
.VERSION  2.0   08-Dec-1994: Take care of newlines embedded in strings,
			Take also care of embedded notes.
.VERSION  2.1   23-Dec-1994: Allow compressions
.VERSION  2.2   04-Feb-1995: Equinox directly from TABINFO
.VERSION  2.3   21-Aug-1995: Strict order XTENSION BITPIX NAXIS NAXIS1
    				NAXIS2 PCOUNT GCOUNT TFIELDS
.VERSION  2.4   26-Oct-1995: EPOCH was not properly edited...
.VERSION  2.5   24-Feb-1997: Adapt the label when array field
.VERSION  2.6   25-Mar-1997: -m -t options
.VERSION  2.7   18-Mar-1999: Be sure option -m 0 is OK
.VERSION  2.8   15-Apr-1999: Convert E/D format to Uppercase E !!!
.VERSION  2.8   26-May-1999: Use the OGIP LONGSTRN convention
.VERSION  2.9   09-Nov-2000: Pb with Linux fixed
.VERSION  2.91  17-Jan-2001: Accept files with .gz or .Z as input
.VERSION  2.92  11-Jun-2004: Bug (crashed for non-existing file) fixed
.VERSION  3.0   27-Jul-2005: Take care of comment lines, empty lines.
.VERSION  3.1   08-Dec-2005: Adjust length of last field
.VERSION  3.2   02-Jun-2007: Cosmetics (-Wall)
.VERSION  3.21  28-Aug-2007: no more blocked ...
.VERSION  3.22  17-Jul-2011: No error message for file in subdir.
.VERSION  3.23  22-Nov-2011: Accept file dir/./file
.VERSION  3.24  05-Sep-2012: Problem with NULL as dashes
.VERSION  3.3   22-Mar-2013: 64-bit
.COMMENTS       Transforms a set of files + ReadMe into FITS tables.
	The following global (environment) variables are used:
	toFITSr = recursivity counter
	toFITSd = 0 (only headers, ascii) / 1 (hdu + data) / 2 (only headers)
	toFITSh = 0 (no SIMPLE header) / 1 (write SIMPLE header)
	toFITSm = 0 (no FILEMARK) / 1 (write FILEMARK)
---------------*/

#define VERSION "3.3"	/* The Version Number */

#define EMBEDDED_NOTE    0	/* 1 to edit notes as embedded comments \
				   in FITS output			*/

#ifndef DEBUG
#define DEBUG 0
#endif

#ifndef MAXPATHLEN
#define MAXPATHLEN	256
#endif

#include <catstd.h>
#include <filutil.h>
#include <string.h>	/* Standard String manipulations */
#include <fcntl.h>	/* File Control definitions	 */
#include <syslog.h>	/* When running by Root		 */
#include <unistd.h>
#include <stdio.h>
#ifdef __MSDOS__
#include <sys\stat.h>
#else
#include <sys/stat.h>
#endif
#ifndef O_BINARY
#define O_BINARY 0
#endif
		/*--------List of the file not transformed to FITS------*/
static char *nofits[] = { 
    ".tex", ".sty", ".dvi", ".ps", ".doc", ".pdf", 
    ".htm", ".html", 
    ".tar", ".tgz", ".exe", ".fgz", 
    ".fih", ".fis",
    "Makefile",
    "/*", 
};

#define MAX(x,y)        ((x) >= (y) ? (x) : (y))
#define MIN(x,y)        ((x) >= (y) ? (y) : (x))
#define ITEMS(a)        sizeof(a)/sizeof(a[0])

#if DEBUG
static char inbyte;
#endif

static char *FITSenv[] = {		/* Environment Variables */
#if O_BINARY		/* MSDOS: Env. variables in UPPERCASE */
    "TOFITSR",		/* Recursivity indicator */
    "TOFITSH",		/* 1 means "write SIMPLE */
    "TOFITSD",		/* 1 means "write data"	 */
    "TOFITSM",		/* 1 means "write FILEMARK */
#else
    "toFITSr",		/* Recursivity indicator */
    "toFITSh",		/* 1 means "write SIMPLE */
    "toFITSd",		/* 1 means "write data"	 */
    "toFITSm",		/* 1 means "write FILEMARK */
#endif
};
#define toFITSr FITSvar[0]
#define toFITSh FITSvar[1]
#define toFITSd FITSvar[2]
#define toFITSm FITSvar[3]

static char FITSvar[] = { 0, 1, 1, 0 };
static FILE *fout ;		/* Output file		*/
static char *ddat;		/* Directory with data	 */
static char *intro;		/* Name of Description File */
static char cwd[MAXPATHLEN];	/* Working directory	*/
static  int oargc; 		/* Original Arguments	*/
static char **oargv;		/* Orignal arguments	*/

static CATINFO *tinfo;
static char *opened_file;	/* Output file	 	*/
static char *SIMPLE_file;	/* The SIMPLE header	*/
static char *zip_pgm[2];	/* Compress pgm + args	*/
static char *zip_ext[] = {".Z", ".z", ".gz" };
static  int fop    = 0;		/* Format Options	*/
static long maxrecs = -1 ;	/* Maximal output/file	*/
static long saxis2  = -1 ;	/* Position of NAXIS2	*/
static long naxis2  = -1 ;	/* Actual      NAXIS2	*/

static char verbop = DEBUG;	/* Verbose option       */
static char ow = 0;  		/* -w options:		*/
#define WOP_fih		2	/* 	-w.fih		*/
#define WOP_fit		4	/* 	-w.fit		*/
static char ofih = 0;  		/* -fih   option	*/
static char ou   = 0;		/* -u option		*/

static int errors = 0;  /* Errors counter       */

static char usage[] = "\
Usage: tofits [-v] [-z|Z] [-u[1|2]] [-w[.fit|.fih]] [-fih] [-noh] [-nodat[a]]\n\
       [-d dir] [-i file] [-o file] [-h file] [-m max] [[-t desc] file]\n\
" ;
static char thelp[] = "\
      -v: verbose option\n\
      -Z: compress the result via compress(1) [can't be combined with -o]\n\
      -z: compress the result via gzip(1)     [can't be combined with -o]\n\
     -u1: generates as FITS comments the SI values of the units\n\
     -u2: generates both litteral and SI values of the units\n\
    -fih: generates the FITS file from a SIMPLE.fit + *.fih + data files\n\
    -noh: don't generate the SIMPLE header\n\
  -nodat: generates only the headers (fully blocked to 2880 bytes)\n\
 -nodata: generates only the headers (as a single ascii file with linefeeds)\n\
  -m max: specifies a maximal number of data records (default: all)\n\
  -d dir: specifies the directory containing the data files\n\
 -i file: specifies the file with description (default 'ReadMe', - for stdin)\n\
 -h file: specifies the file containing the SIMPLE header (default generated)\n\
 -o file: specifies the output file (stdout by default; no compression)\n\
 -t desc: asks to issue a file with tabular structure identical to desc\n\
  -w.fit: generates a new FITS file (extension .fit) for each data file\n\
  -w.fih: generates a header file (extension .fih) for each data file\n\
    file: . for no file at all (only the SIMPLE), - for stdin\n\
          default all data files described in 'ReadMe'\n\
          If 'file' is a directory, option -d 'file' is assumed.\n\
" ;

int gethost();	/* In domain.c */
void setpgm();	/* In error.   */

/*==================================================================
		Internal Routines
 *==================================================================*/

static void cd(dir)
/*++++++++++++++++
.PURPOSE  Internal change of directory. Be sure to keep relative from initial
		directory...
.RETURNS  ---
-----------------*/
  char *dir;	/* IN: New Directory */
{
    if (ddat && *cwd && (*dir != dirsep)) {
	if (chdir(cwd)) perror(cwd);
    }
    if (chdir(dir)) perror(dir), 
    	terminate("Can't find the data directory");
    if (ddat) free(ddat);
    ddat = strdup(dir) ;
}

static void start_out(pgm)
/*++++++++++++++++
.PURPOSE  Start the compression program.
.RETURNS  ---
-----------------*/
  char *pgm;	/* IN: Compression utility */
{
  int fno ;
  FILE *fdopen();
  for (fno=0; fno<100; fno++)
    if (fileno(fout) > 2) terminate ("Can't compress with -o output") ;
    /* Define the output as a Pipe if compression required */
    zip_pgm[0] = pgm ;
    fno = open_pipe(zip_pgm, 1);
    if (fno < 0) terminate ("Can't start the compressing exec.");
    fout = fdopen(fno, WMODE);
    saxis2 = -1 ;			/* Can't seek */
}

static int System(command)
/*++++++++++++++++
.PURPOSE  Internal "system" call: stenv the current variable values and
		execute command
.RETURNS  system status
-----------------*/
  char *command;	/* IN: Command to execute */
{
  char buf[16*4]; 
  int i; char *p ;
    for (p=buf, i=0; i<ITEMS(FITSenv); i++, p += strlen(p)+1) {
    	sprintf(p, "%s=%d", FITSenv[i], FITSvar[i]);
    	putenv(p);
    }
    if ((i = system(command))) perror(command);
    return(i);
}

static void help()
/*++++++++++++++++
.PURPOSE  Detailed help
.RETURNS  -
-----------------*/
{
  int i;
    fprintf(stderr, "%s", usage) ;
    fprintf(stderr, "%s", thelp) ;
    fprintf(stderr,  "Files with names ending by one of the following ");
    fprintf(stderr,  "suffixes are not converted:");
    for (i=0; i < ITEMS(nofits); i++) {
	if ((i%9) == 0) fprintf(stderr, "\n");
	fprintf(stderr, "%-8s", nofits[i]);
    }
    fprintf(stderr, "\n");
}

static int match_nofits(filename)
/*++++++++++++++++
.PURPOSE  Verify a file is in list of non-fits files
.RETURNS  1 (in nofits) / 0
-----------------*/
  char *filename;	/* IN: the file name to examine */
{
  int i; char *e, *f, *p;
    i = strbloc(filename, dirsep); if (i < 0) i = 0; 
    f = filename+i;		/* basename	*/
    i = strbloc(f, '.'); if (i < 0) i = 0; 
    e = f+i;			/* extension 	*/
    for (i=0; i<ITEMS(nofits); i++) {	/* Compare case-INsensitive */
    	p = *nofits[i] == '.' ? e : f;
    	if (studiff(p, nofits[i]) == 0) return(1);
    }
    return(0);
}

static void open_out(name, ext)
/*++++++++++++++++
.PURPOSE  Open an output file with provided extension 
.RETURNS  Opened file / NULL
.REMARKS  Previously opened file is closed;
	The directory is known as cwd.
-----------------*/
  char *name;	/* IN: the data file name ; NULL to close only */
  char *ext;	/* IN: Extension or NULL (see open_ext) */
{
  char *p;
    if (fileno(fout) > 2) { 
    	fit_flush(fout, -1L);
    	if (ferror(fout)) { 
    	    perror(opened_file);
    	    terminate("Error writing FITS output");
    	}
    	fclose(fout); 
    	fout = stderr; 
    	saxis2 = -1 ;		/* Can't seek..	*/
	if (zip_pgm[0]) wait_pipe(0), 
	    zip_pgm[0] = (char *)0 ;
    }
    if (!name) return;		/* No file to open at all */
    p = filename(cwd, name, ext);
    fout = fopen(p, "w+");
    if (!fout) perror(p), terminate((char *)0);
    saxis2 = 0 ;		/* Can   seek..	*/
    opened_file = name;
}

static char *ed_lim(format, n)
/*++++++++++++++++
.PURPOSE  Edit the limit (0,1,2)
.RETURNS  Edited limit (min or max or NULL); starts by ! if Exclusive limit.
-----------------*/
  COLINFO *format;	/* IN: Fortran Format */
  int n;		/* IN: 0=lower lim., 1=upper lim., 2=NULL */
{
  static char fmt[12], buf[60];
  double *pd; int8 *pl; char *p;

    if (!format) return((char *)0);
    if (format->len == 0) return((char *)0);
    fmt[0] = fmt[1] = 0;
    switch(n) {
      case 0:	/* lower */
      	if (!(format->elims & F_IMIN)) return((char *)0);
      	if (format->elims & F_XMIN) fmt[0] = '!';
      	break;
      case 1:	/* upper */
      	if (!(format->elims & F_IMAX)) return((char *)0);
      	if (format->elims & F_XMAX) fmt[0] = '!';
      	break;
      default:
      	n = 2;
      	if (format->elims & F_REQD) {
      	    if (format->elims&F_NULL) break;
      	    return((char *)0);	/* no NULL allowed */
     	}
     	else return(format->ff == 'A' ? (char *)0 : "");
    }
    p = fmt + strlen(fmt);
    switch(format->ff) {
      case 'E': case 'D': 
      	sprintf(p, "%%%d.%dE", format->len, format->len - format->odp -1);
      	break;
      case 'I':
#if int8l == 1
      	sprintf(p, "%%%dld", format->len);
#else
	sprintf(p, "%%%dlld", format->len);
#endif
      	break;
      default:
      	sprintf(p, "%%%d.%d%c", format->len, format->len - format->odp -1,
      	    tolower(format->ff));
    }
    if (format->elims&F_DBLE) {
      	pd = (double *)format->lims;
    	sprintf(buf, fmt, pd[n]);
    } else {
      	pl = (int8 *)format->lims;
    	sprintf(buf, fmt, pl[n]);
    }
    return(buf);
}

static void sec_cleanotes(sec)
/*++++++++++++++++
.PURPOSE  Clean the notes (the flag indicating the notes are edited)
.RETURNS  -
-----------------*/
  SECTION *sec;	/* Starting section to look at */
{
  SECTION *ps;
    ps = sec; 
    if (ps->class != SEC_Note) ps = ps->next;
    while(ps) {
    	if (ps->class != SEC_Note) break;
    	ps->used &= ~SECNOTE_edited;
    	ps = ps->next;
    }
}

#if 0
static FILINFO *edited_filenote(T, note)
/*++++++++++++++++
.PURPOSE  Find out for which file the designated note is first edited
.RETURNS  The file which contains the note / NULL
-----------------*/
  CATINFO *T;	/* IN: Various details about position of description */
  SECTION *note;/* IN: The note to look at */
{
  FILINFO *pf; int i;
    for (pf=T->afinfo, i=T->nfinfo; --i>=0; pf++) {
    	if (pf->note == note) return(pf);
    }
    return((FILINFO *)0);
}

static COLINFO *edited_note(T, note)
/*++++++++++++++++
.PURPOSE  Find out for which file the designated note is first edited
.RETURNS  The file which contains the note / NULL
-----------------*/
  TABINFO *T;	/* IN: Format of the table */
  SECTION *note;/* IN: The note to look at */
{
  COLINFO *pf; int i;
    for (pf=T->afmt, i=T->nfmt; --i>=0; pf++) {
    	if (pf->note == note) return(pf);
    }
    return((COLINFO *)0);
}
#endif

/*==================================================================
		FITS-related output
 *==================================================================*/
static int prMARK(name)
/*++++++++++++++++
.PURPOSE  Write out a FILEMARK extension
.RETURNS  Number of lines
-----------------*/
  char *name;	/* IN: Name of next file */
{
  int lines;
    if (verbop) fprintf(stderr, "\t(FILEMARK)\n");
#if DEBUG
    inbyte = getchar();
#endif
    lines = 0;
    lines += fit_prka(fout, "XTENSION", "FILEMARK", 
    	"indicates the END of current FITS file");
    lines += fit_prka(fout, "EXTNAME", name, "Name of next FITS file");
    lines += fit_prcom(fout, "END", (char *)0);
    fit_flush(fout, -1L);
    return(lines);
}

static int prSIMPLE(argc, argv)
/*++++++++++++++++
.PURPOSE  Write out the SIMPLE header
.RETURNS  Number of lines
-----------------*/
  int argc;	/* IN: Remaining arguments */
  char **argv;	/* IN: Remaining arguments */
{
  char buf[71], *date(), *getenv(), *p, *e, x;
  int lines;
    if (verbop) fprintf(stderr, "\t(SIMPLE)\n");
    lines = 0;
    e = buf + sizeof(buf) - 1;	/* End of buffer */
#if DEBUG
    inbyte = getchar();
#endif
    lines += fit_prkn(fout, "SIMPLE", "T", "Standard FITS Format");
    lines += fit_prki(fout, "BITPIX", 8, "Character data");
    lines += fit_prki(fout, "NAXIS", 0, "No Image --- just extension(s)");
    lines += fit_prkn(fout, "EXTEND", "T", "There are standard extensions");
    /*
    lines += fit_prkn(fout, "BLOCKED", "T", 
    		"File may be blocked (obsolete, but who knows?)");
    */
    lines += fit_prka(fout, "ORIGIN", "CDS", 
    		"File generated at CDS, Strasbourg, France");
    sprintf(buf, 
    	"                                 (tofits, Version %s)", VERSION);
    lines += fit_prcom(fout, buf, "");
    lines += fit_prcom(fout, 
    	"             e-mail:    question@simbad.u-strasbg.fr", "");
    strcpy(buf, "ARG=");			/* Write out the Arguments */
    p = buf + strlen(buf); x = '\'';
    while(--argc > 0) {
    	if ((p+strlen(*++argv)) >= &(buf[sizeof(buf)-6])) break;
    	*(p++) = x; x = ' ';
    	strcpy(p, *argv); p += strlen(p);
    }
    *(p++) = '\''; *p = 0;
    if (argc>0) strcpy(p, "...");
    if ((p-buf) > 5) lines += fit_prcom(fout, "COMMENT", buf);
    lines += fit_prka(fout, "LONGSTRN", "OGIP 1.0", 
        "Long string convention (&/CONTINUE) may be used") ;
    sprintf(buf, "Written on %s (GMT)", date(':'));	/* Write out the Date */
    lines += fit_prka(fout, "DATE", date('F'), buf);
    p = getenv("USER"); if (!p) p = "(nobody)";
    sprintf(buf, "                             by: %s@", p);
    p = buf + strlen(buf) ;
    gethost(p, e-p) ;
    lines += fit_prcom(fout, buf, "");
    return(lines);
}

static int countRecords(finfo, file, max)
/*++++++++++++++++
.PURPOSE  Print out the Data file
.RETURNS  Number of data lines
.REMARKS  Empty lines are skipped only in .dat files
-----------------*/
  FILINFO *finfo;	/* IN: The file with all details */
  char *file;		/* IN: Actual file name */
  int  max;		/* Stop if max reached  */
{
  FILE *f; char *buf, headchar=0;
  TABINFO *pt; 
  int lbuf, len, headlines=0, skipped_lines=0;
  long lines=0;

    if (!(f = fzopen(file))) {
	fprintf(stderr, "#***tofits([%s]%s): ", finfo->dfile, file);
	perror("");
	return(-1);
    }
    lbuf = finfo->reclen+4; 
    buf  = malloc(lbuf); 
    if ((pt = finfo->tabform)) {
	/* Maybe some comment / headlines should be ignored... */
	headlines = pt->headlines;
	headchar  = pt->headchar;
    }

    while (fgets(buf, lbuf, f)) {
	if (headlines>0) {	/* V3.0 added ---- */
	    if (verbop) fprintf(stderr, "....[comment]%s", buf);
	    skipped_lines++;
	    --headlines; 
	    continue; 
	}
	if (headchar) { 	/* V3.0 added ---- */
	    if (*buf == headchar) {
	        if (verbop) fprintf(stderr, "....[comment]%s", buf);
	        skipped_lines++;
		continue;
	    }
	    headchar = 0;	/* No more headlines */
	}
    	len = strline(buf);
	if (len == 0) {		/* Empty line... */
	    skipped_lines++; 
	    continue; 
	}
	lines++;
	if ((max>0) && (lines==max)) /* No need to count more*/
	    break;
    }
    if (verbop && skipped_lines) fprintf(stderr, 
	    "#...count(%s): skipped_lines=%d\n", file, skipped_lines);
    fzclose(f); free(buf);
    return(lines);
}

static int prData(finfo, input)
/*++++++++++++++++
.PURPOSE  Print out the Data file
.RETURNS  Number of data lines
.REMARKS  Empty lines are skipped only in .dat files
	There is a need to translate the formats 'D' 'E'
-----------------*/
  FILINFO *finfo;	/* IN: The file with all details */
  FILE *input;		/* IN: The opened file, or NULL	 */
{
  static char trt[3*256] ;	/* Conversion tables */
  FILE *f; char *buf, *ebuf, *p, *amod, trm, headchar=0;
  TABINFO *pt; COLINFO *pf, *ef ;
  int i, lbuf, len, mode, headlines, keep_blank_lines, skipped_lines;
  long bytes, modulo, lines;
  int blanking=0, *blank_minus=0;  /* Added V3.24 */

    lines = headlines = skipped_lines = 0;
    mode = fit_getmode();
    if (verbop) {
	fflush(stdout);
	fprintf(stderr, "\t(mode=%d, %dx%d-byte records from file '%s')\n", 
    	    mode, finfo->records, finfo->reclen, finfo->dfile);
    }
#if DEBUG
	inbyte = getchar();
#endif
    if (input) f = input ;
    else f = fzopen(finfo->dfile);
    if (!f) {
	fprintf(stderr, "****Can't open data file: %s\n", finfo->dfile) ;
	terminate((char *)0);
    }
    lbuf = finfo->reclen+4; 
    amod = malloc(2*lbuf); 
    buf  = amod + lbuf ;
    ebuf = buf + finfo->reclen;
    i = strloc(finfo->dfile, '.') ;
    /* V3.0: keep_blank_lines is zero when file is a data file  */
    keep_blank_lines = strdiff(finfo->dfile+i, ".dat");

    /* Are there conversions ? If yes, assign amod as a translator */
    memset(amod, 0, finfo->reclen) ;
    if ((pt = finfo->tabform)) {
	for (pf=pt->afmt, ef=pf + pt->nfmt; pf<ef; pf++) {
	    if (pf->ff == 'D') 
		for (i = pf->b0; i <= pf->b1; i++) amod[i] = 2 ;
	    else if (pf->ff == 'E') 
		for (i = pf->b0; i <= pf->b1; i++) amod[i] = 1 ;
	    else if ((pf->elims&(F_CHAR|F_NULL))==(F_CHAR|F_NULL)) /* V3.24 */
		blanking++;
	}
	/* Maybe some comment / headlines should be ignored... */
	headlines = pt->headlines;
	headchar  = pt->headchar;
	/* V3.24: Convert --- into blanks: indicate in the array "blanking"
	 * the fist/last byte to check for a conversion info blanks */
	if (blanking) {
	    blanking = (2*blanking+1); 
	    blank_minus = (int *)malloc(sizeof(int)*blanking);
            memset(blank_minus, -1, sizeof(int)*blanking);
	    for (pf=pt->afmt, i=0; pf<ef; pf++) {
		if ((pf->elims&(F_CHAR|F_NULL))!=(F_CHAR|F_NULL)) continue;
		blank_minus[i++] = pf->b0;
		blank_minus[i++] = pf->b1;
	    }
	}
    }
    /* Set trm to 1 if conversions are required */
    /* TODO: Convert '---' to BLANKS --- */
    for (trm = 0, i = finfo->reclen; (--i >= 0) && (trm == 0); trm = amod[i]) ;
    if (trm) for (i=0; i<256; i++) {
       trt[i] = i ;
       trt[i+256] = i ;
       trt[i+512] = i ;
    }
    trt[256+'e'] = 'E' ;
    trt[512+'d'] = 'D' ;
    trt[512+'e'] = 'D' ;
    trt[512+'E'] = 'D' ;

    bytes = 0; modulo = 1;
    while (fgets(buf, lbuf, f)) {
	if (headlines>0) {	/* V3.0 added ---- */
	    if (verbop) fprintf(stderr, "....[comment]%s", buf);
	    skipped_lines++;
	    --headlines; 
	    continue; 
	}
	if (headchar) { 	/* V3.0 added ---- */
	    if (*buf == headchar) {
	        if (verbop) fprintf(stderr, "....[comment]%s", buf);
	        skipped_lines++;
		continue;
	    }
	    headchar = 0;	/* No more headlines */
	}
    	len = strline(buf);
	if (len == 0) {		/* V3.0 added ---- */
	    if (keep_blank_lines) ;
	    else { skipped_lines++; continue; }
	}
	if (mode) {	/* Must block lines */
	    /* V3.24: Blank the minuses if necessary */
	    if (blank_minus) 
	      for (blanking=0; blank_minus[blanking]>=0; blanking++) {
		i = blank_minus[blanking++]; 
		while ((i<=blank_minus[blanking]) && (buf[i] = ' ')) i++;
		if ((i<blank_minus[blanking]) 
			&& (buf[i] == '-') && (buf[i+1] == '-')) {
		    while ((i<=blank_minus[blanking]) && (buf[i] == '-'))
			buf[i++] = ' ';
		}
	    }
	    if (trm) for (i=0; i<len; i++) 
		buf[i] = trt[(amod[i]<<8)+((unsigned char *)buf)[i]] ;
	    p = buf + len;
	    while (p < ebuf) *(p++) = ' ';
	    bytes += fwrite(buf, 1, finfo->reclen, fout);
	}
	else {
    	    bytes += fwrite(buf, 1, len, fout);
	}
    	if (ferror(fout)) break;
    	lines++;
	if (lines == maxrecs) break ;
	if (verbop) {
	    if (lines == 10) modulo = 10;
	    else if (lines == 100) modulo = 100;
	    else if (lines == 1000) modulo = 1000;
	    else if (lines == 10000) modulo = 5000;
	    else if (lines == 100000) modulo = 10000;
	    if (lines%modulo == 0) fprintf(stderr, "\r\t(lines: %6ld", lines);
	}
    }
    if (verbop) {
	fprintf(stderr, "\r\t(lines=%ld, bytes=%ld", lines, bytes);
	if (skipped_lines) fprintf(stderr, 
	    ", skipped_lines=%d", skipped_lines);
	fprintf(stderr, ")\n");
    }

    if (lines != naxis2) {
	fflush(fout) ;
	if (saxis2 >= 0) saxis2 = fseek(fout, saxis2, 0);
	if (saxis2 >= 0) {
	    fit_prki(fout, "NAXIS2", lines, "Number of records");
	    if (fseek(fout, 0L, 2) < 0)	/* Go to End of created file	   */
	    	perror(opened_file),
	    	terminate("Can't properly seek in output") ;
	}
	else fprintf(stderr, 
    	    "****File %s: %ld lines copied (%ld assumed)****\n", 
	    finfo->dfile, lines, naxis2);
    }
    fit_flush(fout, bytes); 	/* Fill the last block with blanks */

    free(amod) ;
    if (blank_minus) free(blank_minus);	/* V3.24 */
    	
    return(lines);
}

static int prTitle(text)
/*++++++++++++++++
.PURPOSE  Print out the Title
.RETURNS  Number of lines
-----------------*/
  char *text;	/* IN: Text of title including number, author, etc */
{
  char *p, *e, *n, x;
  int lines, i, m, k=0;
  char buf[20] ;	/* Keyword */

    /* k indicates 1=title 2=author 4=reference 8=bibcode */
    /* m is a counter increasing when several papers	  */
    m = 0 ;
#if DEBUG>1
    printf("prTitle: %s", text); inbyte = getchar();
#endif
    lines = 0;
    for (p=text; isgraph(*p); p++) ; x = *p; *p = 0;
    lines += fit_prka(fout, "CDS-CAT", text, 
    	"Catalogue designation in CDS nomenclature");
    *p = x;
    n = p + strloc(p, '\n'); while (isspace(*p)) p++;
    if (p < n) {	/* There is a "Short Title" */
    	e = p + strnline(p, n-p);
    	x = *e; *e = 0;
    	lines += fit_prcom(fout, "COMMENT", p);
    	*e = x;
    }
    for (p=n; *p; p=n) {
#if DEBUG>1
    	printf("prTitle: remaining bytes=%d\r%s", strlen(p), p); 
    	inbyte = getchar();
#endif
    	i = strloc(++p, '\n'); n = p+i;
    	if (match_rule(p)) continue;	/* Just ignore rules */
    	if (isgraph(*p)) {		/* Title */
    	    while(*n) {
    	    	if (!isgraph(n[1])) break;
    	    	++n; n += strloc(n, '\n');
    	    }
    	    x = *n; *n = 0;
    	    lines += fit_prcom(fout, "COMMENT  Title: ", p);
    	    *n = x;
	    if (k != 1) ++m ;	/* Counter for # of titles */
	    k = 1 ;		/* Indicates title, no author / ref yet */
    	    continue;
    	}
    	while(*p == ' ') p++;
    	for (e=n; isspace(*e); e--) ; x = *++e; *e = 0;
    	if (*p == '=') { char *e1, x1;
	    if (k&8) ++m ; k |= 8 ;
	    strcpy(buf, "BIBCODE") ;
	    if (m > 1) sprintf(buf+6, "%d", m) ;
	    for (e1=p; isgraph(*e1); e1++) ; x1 = *e1; *e1 = 0;
	    lines += fit_prka(fout, buf, p+1, 
    		"19-digit SIMBAD/NED/ADS BibCode");
	    *e1 = x1;
	}
    	else if (*p == '<') {
	    if (k>4) ++m, k=0 ; k |= 4 ;
    	    *e = x; p++; 
    	    e = p + strloc(p, '>'); 	/* End of Reference */
	    n = e + strloc(e, '\n');	/* May be on several lines */
	    x = *e; *e = 0;
	    strcpy(buf, "REFERENC") ;
	    if (m > 1) sprintf(buf+6, "%d", m) ;
    	    lines += fit_prka(fout, buf, p, (char *)0);
    	}
    	else {				/* Authors, locate lines */
	    *e = x ;
	    if (k>2) ++m, k=0 ; k |= 2 ;
	    strcpy(buf, "AUTHOR") ;
	    if (m > 1) sprintf(buf+6, "%d", m) ;
    	    while(*n) {
		for (i=1; n[i] == ' '; i++) ;
		if (i <= 1) break ;
		if (!isalpha(n[i])) break ;
    	    	++n; n += strloc(n, '\n');
    	    }
    	    x = *n; *n = 0;
	    lines += fit_prka(fout, buf, p, (char *)0);
    	    *n = x;
    	    continue;
	}
    	*e = x;
    }
    return(lines);
}

static int prForm(finfo, records)
/*++++++++++++++++
.PURPOSE  Print out the "Byte-by-Byte description" (set of column formats)
.RETURNS  Number of lines
-----------------*/
  FILINFO *finfo;	/* IN: The file with all details */
  long records ;	/* IN: Number of records	 */
{
  TABINFO *T; COLINFO *pf; SECTION *ps; 
  int i, j, k, l, m, lines;
  char buf[256], kbuf[32], *p, *plabel;
  static char *tit[] = {"TAMIN", "TAMAX", "TBNUL"};
  static char *com[] = {
  	"Allowed minimal value", 
  	"Allowed maximal value", 
  	"NULL (undefined) value"
  };

    lines = 0;
    T = finfo->tabform;

    if (verbop) fprintf(stderr, "\t(XTENSION %s)\n", finfo->dfile);
#if DEBUG
	inbyte = getchar();
#endif
    if (T) {	/* V3.1: verify reclen compatible */
	for (pf=T->afmt+T->nfmt, l=0; --pf>=T->afmt; ) {
	    if (l <= pf->b1) l = pf->b1 + 1;
	}
	if (l > finfo->reclen) {	/* Must append blanks */
	    if (verbop) fprintf(stderr,
	      "++++prForm: file %s: reclen increased from %d to %d\n", 
	      finfo->dfile, finfo->reclen, l);
	    finfo->reclen = l;
	}
	else if (l < finfo->reclen) {	/* Should truncate ??? */
	    if (verbop) fprintf(stderr, 
	      "++++prForm: file %s: reclen=%d > %d\n",
	      finfo->dfile, finfo->reclen, l);
	}
    }

    lines += fit_prka(fout, "XTENSION", "TABLE", "Ascii Table Extension");
    lines += fit_prki(fout, "BITPIX", 8, "Character data");
    lines += fit_prki(fout, "NAXIS", 2, "Simple 2-D matrix");
    lines += fit_prki(fout, "NAXIS1", finfo->reclen, 
    	"Number of bytes per record");
    if (saxis2 >= 0) saxis2 = ftell(fout) ;
    naxis2 = records >= 0 ? records : finfo->records ;
    lines += fit_prki(fout, "NAXIS2", naxis2, "Number of records");
    lines += fit_prki(fout, "PCOUNT", 0, "Get rid of random parameters");
    lines += fit_prki(fout, "GCOUNT", 1, "Only one group (isn't it obvious?)");

	/* Count the number of TFIELDS, and edit it */
    if (T) {		/* Edit also EQUINOX and EPOCH */
	for (i = T->nfmt, pf = T->afmt, m=0; --i >= 0; pf++) m += pf->rep ;
    	lines += fit_prki(fout, "TFIELDS", m, 
	    "Number of data fields (columns)");
	p = finfo->equinox;
	if (!p) p = T->equinox ;
    	if (p) {	/* Found EQUINOX, edit it */
    	    strcpy(buf, "Equinox of coordinates");
    	    if (isalpha(*p)) sprintf(buf+strlen(buf), " (%c system)", *p),  p++;
	    for (i=0, k=0; isgraph(*p); i++, p++) {	/* k != 0 if . exists */
		kbuf[i] = *p; 
		if (*p == '.') k++;
	    }
	    if (!k) kbuf[i++] = '.';
	    kbuf[i] = 0;
    	    lines += fit_prkn(fout, "EQUINOX", kbuf, buf);
    	}
	p = finfo->epoch;
	if (!p) p = T->epoch ;
    	if (p) {		/* Found EPOCH  , edit it */
    	    strcpy(buf, "Epoch   of coordinates");
    	    if (isalpha(*p)) sprintf(buf+strlen(buf), " (%c system)", *p),  p++;
	    for (i=0, k=0; isgraph(*p); i++, p++) {	/* k != 0 if . exists */
		kbuf[i] = *p; 
		if (*p == '.') k++;
	    }
	    if (!k) kbuf[i++] = '.';
	    kbuf[i] = 0;
    	    lines += fit_prkn(fout, "EPOCH" , kbuf, buf);
    	}
    }
    else lines += fit_prki(fout, "TFIELDS",1, 
	    "Just a single column, i.e. Text file");
    
    lines += fit_prka(fout, "EXTNAME", finfo->dfile, finfo->text);

    if ((ps = finfo->note)) {	/* Edit notes associated to file */
    	lines += fit_prcom(fout, "COMMENT", ps->header);
    	l = ps->ebody - ps->abody;
    	if (ps->abody && (l>0)) lines += fit_prt(fout, "", ps->abody, l);
    }

    if (!T) {		/* No description; assume text */
    	lines += fit_prki(fout, "TBCOL1", 1, "Start column");
    	sprintf(buf, "A%d", finfo->reclen);
    	lines += fit_prka(fout, "TFORM1", buf, "Fortran Format");
    	lines += fit_prka(fout, "TTYPE1", "Text", "Just free-format text");
    	return(lines);
    }

    		/* Edit now the columns */
    lines += fit_prt(fout, "", rule('='), 80);
    if (ps) sec_cleanotes(ps);
    for (i=T->nfmt, pf=T->afmt, j=0; --i>=0; pf++) for (m=pf->rep; --m>=0; ) {
    	j++;
    	sprintf(kbuf, "TBCOL%d", j);
    	sprintf(buf, "============== Start column +%d", pf->b0);
    	lines += fit_prki(fout, kbuf, pf->b0+1, buf);
    	p = unitex(pf->unit, ou);
    	if (*p) {	/* Units */
    	    sprintf(kbuf, "TUNIT%d", j);
    	    strcpy(buf, "Unit: "); strcat(buf, p);
    	    lines += fit_prka(fout, kbuf, pf->unit, buf);
    	}
    	p = fmt_ed(pf); while (isdigit(*p)) p++;	/* Forget factor */
    	sprintf(kbuf, "TFORM%d", j);
    	lines += fit_prka(fout, kbuf, p, "Fortran Format");
    	kbuf[1] = 'D', kbuf[2] = 'I', kbuf[3] = 'S', kbuf[4] = 'P';
    	if (*p != 'A') lines += fit_prka(fout, kbuf, p, 
    	    "Display Format for Binary Tables");
    	sprintf(kbuf, "TTYPE%d", j);
	if (pf->rep>1) {
	    plabel = buf;
	    sprintf(plabel, "%s:%d/%d", pf->label, pf->rep-m, pf->rep) ;
	}
	else plabel = pf->label ;
    	lines += fit_prka(fout, kbuf, plabel, pf->text);
    			/* Edit the TAMIN/TAMAX/TNULL Keywords */
    	for (k=0; k<=2; k++) if((p=ed_lim(pf, k))) {
            sprintf(kbuf, "%s%d", tit[k], j);
    	    if (*p) {	/* Check for EXCLUSIVE Limits */
    	    	strcpy(buf, com[k]);
    	    	if (*p == '!')	/* Indicates EXCLUSIVE limit */
    	    	    p++, strcat(buf, " EXCLUSIVE (never reached)");
    	    	lines += fit_prkn(fout, kbuf, p, buf);
	    }
    	    else {	/* Issue the relevant number of spaces */
    	    	for (p=buf, l=pf->len; --l>=0; p++) *p = ' ';
    	    	*p = 0;
    	    	lines += fit_prka(fout, kbuf, p, com[k]);
	    }
        }
        		/* Edit the Allowed Character Set 	*/
	if (pf->elims & F_ATRT) {
            sprintf(kbuf, "TASET%d", j);
	    for (k=0, p=buf; k<256; k++) if (pf->lims[k]) *(p++) = k;
	    *p = 0;
	    lines += fit_prka(fout, kbuf, buf, "Allowed character set");
	}
        	/* Edition of Note */
#if EMBEDDED_NOTE
    	if (ps = pf->note) {
    	    if (ps->used & SECNOTE_edited) pfn = edited_note(T, ps);
    	    else pfn = (COLINFO *)0;
    	    if (pfn) {	/* Same note as above */
    	        k = 1+(pfn-T->afmt);
    	        sprintf(buf, "on %s: see note above on Col.#%d (%s)", 
    	            pf->label, k, pfn->label);
    	        lines += fit_prcom(fout, "COMMENT", buf);
	    } else {
    	    	fit_prcom(fout, "COMMENT", ps->header);
    	    	l = ps->ebody - ps->abody;
    	    	if (ps->abody && (l>0)) lines += fit_prt(fout, "",
    			ps->abody, l);
	    }
    	    ps->used |= SECNOTE_edited;
    	}
#endif
    }

#if !EMBEDDED_NOTE		/* Write all notes at the end */
    	/* Write out all notes related to the Table */
    lines += fit_prt(fout, "", rule('='), 80);
    ps = T->section;
    for (++ps, i=0; ps; ps=ps->next, i++) {
    	if (ps->class != SEC_Note) break;
    	lines += fit_prcom(fout, "COMMENT", ps->header);
    	l = ps->ebody - ps->abody;
    	if (ps->abody && (l>0)) lines += fit_prt(fout, "", ps->abody, l);
    }
    if (i) lines += fit_prt(fout, "", rule('='), 80);
#endif
    return(lines);
}

static int prSummary(pf)
/*++++++++++++++++
.PURPOSE  Print out the "File Summary"
.RETURNS  Number of lines written
-----------------*/
  FILINFO *pf;		/* IN: The parameters to print */
{
  char buf[80], *p;
  int lines;
  
    if (!pf) {		/* Introduction */
    	sprintf(buf, "%10s%-15s %5s %8s   %s", "", 
    	    "FileName", "Lrecl", "Records", "Explanations");
    	return(fit_prt(fout, buf, "", 0));
    }
    if (*(pf->dfile) == 0177) return(0);
    p = buf;
    sprintf(p, "%10s%-15s %5d", "", pf->dfile, pf->reclen); 
    p += strlen(buf);
    if (pf->records == 0) sprintf(p, " %8s   ", ".");
    else sprintf(p, " %8d   ", pf->records);
    lines = fit_prt(fout, buf, pf->text, strlen(pf->text));
#if 0
    if (ps = pf->note) {
    	if (ps->used & SECNOTE_edited) pfn = edited_filenote(tinfo, ps);
    	else pfn = (FILINFO *)0;
    	if (pfn) {	/* Same note as above */
    	    sprintf(buf, "See note above on %s", pfn->dfile);
    	    lines += fit_prcom(fout, "COMMENT", buf);
    	    return(lines);
    	}
        /* p = ps->header + strloc(ps->header, ':');
        if (*p) for (++p; isspace(*p); p++) ;
    	sprintf(buf, "COMMENT  on File %s: ", pf->dfile);
    	if (*p) lines += fit_prcom(fout, buf, p), buf[0] = 0;*/
    	fit_prcom(fout, "COMMENT", ps->header);
    	if (ps->abody) lines += fit_prt(fout, "",
    		ps->abody, ps->ebody - ps->abody);
    	ps->used |= SECNOTE_edited;
    }
#endif
    return(lines);
}

static int prIntro(T)
/*++++++++++++++++
.PURPOSE  Print out the "ReadMe" file
.RETURNS  Number of lines
-----------------*/
  CATINFO *T;	/* IN: Various details about position of description */
{
  SECTION *ps, *pc; FILINFO *pf;
  char *p, edit_mask; int i, lines;

    lines = 0; edit_mask = SECNOTE_edited;
    if (verbop) fprintf(stderr, "\t(Intro from %s)\n", T->ffile);
#if DEBUG
	inbyte = getchar();
#endif
    for (ps=T->section[0]; ps; ps=ps->next) switch(ps->class) {
    	case SEC_Title:		/* Write out the designation */
    	    lines += prTitle(ps->abody);
    	    continue;
    	case SEC_BytebyByte:		/* Write out the designation */
    	    edit_mask = SECNOTE_used;		/* Don't issue notes */
    	    p = ps->header; p += strloc(p, ':'); p++;
    	    lines += fit_prt(fout, "COMMENT  Unused Described file: ", 
    	    	p, strlen(p));
    	    continue;
    	case SEC_FileSummary:		/* Write out the designation */
    	    edit_mask = SECNOTE_edited;		/* Issue notes */
    	    lines += fit_prcom(fout, "COMMENT", ps->header);
    	    lines += fit_prt(fout, "", rule('-'), 80);
    	    lines += prSummary((FILINFO *)0);
    	    lines += fit_prt(fout, "", rule('-'), 80);
    	    for (pf=T->afinfo, i=T->nfinfo; --i>=0; pf++) 
    	    	lines += prSummary(pf);
    	    lines += fit_prt(fout, "", rule('-'), 80);
    	    for (pc = ps->next; pc; pc = pc->next) {
    	    	if (pc->class != SEC_Note) break;
    	    }
    	    continue;
    	default:
    	    if (ps->used & edit_mask) continue;
    	    lines += fit_prcom(fout, "COMMENT", ps->header);
    	    for (p=ps->abody; p<ps->ebody; p += i+1) {
    	    	i = strlen(p);
    	    	if (!*p) continue;
    	    	lines += fit_prt(fout, "", p, i);
    	    }
    	    continue;
    	case SEC_End:
    	    lines += fit_prt(fout, "", rule('='), 80);
    	    lines += fit_prcom(fout, "HISTORY", ps->header);
    	    continue;
    }
    	/* Write the "End" */
    lines += fit_prcom(fout, "END", (char *)0);
    fit_flush(fout, -1L);
    return(lines);
}

static void setup()
/*++++++++++++++++
.PURPOSE  Set up what's required for a new file.
.RETURNS  ---
-----------------*/
{
  char buf[40] ;
  int i ;
  
    toFITSr += 1;			/* Recursivity counter	*/
    if (ow&WOP_fit) toFITSd = 1;
    if (ow&WOP_fih) toFITSd = 0;
    if (toFITSm) toFITSh = 0;	/* If FILEMARK no initial SIMPLE */
    fit_setmode(toFITSd);	/* With option "-nodata", simple ASCII file */
    if (verbop) {
	for (i=0; i<ITEMS(FITSenv); i++) fprintf(stderr, 
	     "....Env: %s=%d\n", FITSenv[i], FITSvar[i]);
    }

    if (toFITSr > 1) return ;		/* Setup executed once.	*/

    /* Set up the default variables */
    if (! intro) {
    	intro = "ReadMe";
    	if (!comfile(intro)) intro = "Intro";
    }

    /* When there is a .MakeFITS, execute it; that's end of all	*/
    if (comfile(MakeFITS)) {
#ifdef __MSDOS__
    	sprintf(buf, "%s", MakeFITS);
#else
    	sprintf(buf, "/bin/sh %s", MakeFITS);
#endif
    	exit(System(buf));
    }

    /* Load the "ReadMe" file */ 
    if (ofih) {
    	if (!SIMPLE_file) SIMPLE_file = "SIMPLE.fih"; 
    }
    else {
#if DEBUG
    	printf("....call cad_new(\"%s\", 0x%X)\n", intro, fop); osmtrace(0);
#endif
	if (!tinfo) tinfo = cad_new(intro, fop);	/* Format File */
    	if (!tinfo) terminate("Errors in \"ReadMe\", terminated");
    	if (!tinfo->ffile) tinfo->ffile = intro;
    	if (verbop) cad_check(tinfo);
    }
}

static void convert(file, input)
/*++++++++++++++++
.PURPOSE  Convert a file into FITS
.RETURNS  ---
-----------------*/
  char *file ;	/* IN: The file to convert --- 	*/
  FILE *input;	/* IN: Another file similar...	*/
{
  static FILINFO filinfo ; FILINFO *pf;
  long nrecs ;
  int isFITS, i ;
  char *p;
  
    pf = (FILINFO *)0;
    if (tinfo) {		/* Described  File 	*/
	pf = cad_findfile(tinfo, file);
    	if (!pf) {  		/* Remove the .Z suffix	*/
	    i = strbloc(file, '.');
	    if (i > 0) { 
		p = file + i;
		for (i = 0; i < ITEMS(zip_ext); i++) {
		    if (strdiff(p, zip_ext[i]) == 0) break; 
		}
		if (i < ITEMS(zip_ext)) {	/* Retry without .Z */
		    *p = 0 ;
	    	    pf = cad_findfile(tinfo, file);
	    	    if (!pf) *p = '.' ;
	        }
	    }
        }
    }
    if (!pf) {
    	if (input) { error(file); terminate("File not described") ; }
    	pf = &filinfo, pf->dfile = file ;
    }
	
    if (verbop) fprintf(stderr, "....Dealing with file [%s] %s\n", pf->dfile,
	    file);

    /* Files never converted to FITS: either via match_nofits,
	or file with same name as "ReadMe" file
    */
    p = tinfo ? tinfo->ffile : "";
    if (strdiff(p, pf->dfile) == 0) {	/* It's the 'ReadMe' file */
    	if (verbop) fprintf(stderr, "\t(not converted)\n");
	return;
    }
    /* Files *.dat are accepted */
    if (stumatch("*.dat", file)) ;	/* Accept filename terminated by .dat */
    else if (match_nofits(pf->dfile)) {
    	if (verbop) fprintf(stderr, "\t(not converted)\n");
	fprintf(stderr, "#+++tofits: '%s' can't be converted\n", file);
    	return ;
    }

    isFITS = stumatch("*.fits", pf->dfile);
    if (!isFITS) isFITS = stumatch("*.fit", pf->dfile);
#if DEBUG
        printf("----File: %s; isFITS=%d", pf->dfile, isFITS);
	inbyte = getchar();
#endif

    /* Write first the FILEMARK if necessary */
    if (toFITSm) {
	if (!ow) prMARK(pf->dfile);
	toFITSm = 0;
	toFITSh = 1;	/* Must be followed by a SIMPLE header */
    }
    	
    /* Open the Output FITS File if Necessary */
    if (ow&WOP_fit) {
    	toFITSh = 1;
	open_out(pf->dfile, ".fit");
    }

    /* A FITS file has to be now completely written out */
    if (isFITS) {
	if(verbop) fprintf(stderr, "\t(plain copy of %s)\n", pf->dfile);
#if DEBUG
	inbyte = getchar();
#endif
	if (toFITSd) { 			/* Simple copy */
	    if (!fit_copy(fout, pf->dfile)) errors++,
	    	error("empty file?");
	}
	else fprintf(stderr, "++++Header part of '%s' not extracted\n", 
	    pf->dfile);
	toFITSm = 1;	/* Next file output will require a MARK */
	return ;
    }

    /* Write out the SIMPLE header (except if a FITS file 
    	is to be issued, done previously)
    */
    if (toFITSh) {
	if (ow&WOP_fih) open_out("SIMPLE", ".fih");
    	if (ofih) {		/* Copy SIMPLE_file */
    	    if (verbop) fprintf(stderr, "\t(copy %s)\n", SIMPLE_file);
#if DEBUG
		inbyte = getchar();
#endif
    	    if (!fit_prfile(fout, SIMPLE_file)) errors++, 
    	    	error("Missing file?");
    	}
    	else {
    	    prSIMPLE(oargc, oargv);
    	    prIntro(tinfo);
    	}
    	fit_flush(fout, -1L);
    	toFITSh = 0;
    }
    
    /* Issue the "XTENSION" for ascii tables;
	either from a ".fih" file, or via prForm.
	The variable "i" is set to 1 if we've to
	use such a .fih file.
    */
    p = filename(".", pf->dfile, ".fih");
    i = ofih;
    if (!i) i = (comfile(p) != (char *)0);
    
    if (i) {			/* From .fih 	*/
	if (verbop) fprintf(stderr, "\t(copy %s)\n", p);
#if DEBUG
	inbyte = getchar();
#endif
    	if (!fit_prfile(fout, p)) errors++, error("Missing file?");
    	pf->reclen  = fit_naxis(1);
    	pf->records = fit_naxis(2);
    }
    else {		/* Have to create the Xtension from ReadMe */
    	if (pf == &filinfo) {
    	    errors++; 
    	    fprintf(stderr, "****File not described in %s: %s\n", 
		    tinfo ? tinfo->ffile : "(?ReadMe?)", 
    	    	    pf->dfile);
	}
	if (ow&WOP_fih) open_out(pf->dfile, ".fih");
	/* V3.22: files in sudirectories: pf->records represents the
	 * number of files, not the number of records! */
	if ((pf->flags&F_INDX) && (!input)) {	/* Count #records */
	    pf->records = countRecords(pf, file, maxrecs);
	}
	nrecs = maxrecs ;
	if ((nrecs > 0) && (nrecs > pf->records))
		nrecs = pf->records ;
    	prForm(pf, nrecs);
    	fit_prcom(fout, "END", (char *)0); 	/* Write the "End" */
    }
    fit_flush(fout, -1L);

    /* Now, issue the Data ... */
    if ((toFITSd==1) && maxrecs) { FILE *myFile=0;
	if (strdiff(file, pf->dfile)) {
	    myFile = fzopen(file);
	    if (!myFile) {
		fprintf(stderr, "#***tofits([%s]%s): ", pf->dfile, file);
		perror("");
	    }
	    input = myFile;
	}
	prData(pf, input); 	
	if (myFile) fzclose(myFile);
    }
}

/*==================================================================*/

#define pVal(text)      \
	if (! *++p) {                           	\
	    if (--argc <= 0) errors++, error(text);   	\
	    else    p = *++argv;            		\
	}
/*==================================================================
		Program
 *==================================================================*/
int main(int argc, char **argv)
/*++++++++++++++++
.PURPOSE  Program which concerts to FITS
-----------------*/
{
  char *p, *d, *tparm, *getenv(), *getcwd();
  FILINFO *pf;
  FILE *input ;
  int i, nf;

    setpgm("tofits");
    if (!fout) fout = stdout ;
    oargc = argc; oargv = argv;
    nf = 0 ;	/* Number of files processed */
    tparm = (char *)0 ;

    /* Under ftpd, redirect stderr to 1 */
    /* if (getuid() == 12) { close(2); dup(1) ; } */
    /* fprintf(stderr, "sizeof(cwd)=%d\n", sizeof(cwd)); */
/*
for (i=0; i<=2; i++) fprintf(stderr, "....tofits-main..: fstat(%d)=%d\n", i, fstat(i, &bstat)) ;
*/
    if (!getcwd(cwd, sizeof(cwd))) {
    	perror(*argv);
	strcpy(cwd, ".");
	fprintf(stderr, "++++Can't get cwd, assume \"%s\" [sizeof(cwd)=%d]\n", 
	    cwd, (int)sizeof(cwd)) ;
	/* terminate("Can't get cwd"); */
    }
    
    /* Check first the Environment */
    for (i=0; i<ITEMS(FITSenv); i++) {
    	if ((p = getenv(FITSenv[i]))) FITSvar[i] = atoi(p);
    }
    
    /* Loop on the Arguments */
    while (-- argc > 0) {
	p = *++argv;
	if (*p == '-') switch (*++p) {
	  case 'd' :                   	/* Tell  the Directory 	*/
		pVal("Missing -d Directory");
		cd(p) ;
		continue;
	  case 'o' :                   	/* Tell the output file	*/
	  	pVal("Missing -o(utput) file");
		fflush(fout) ;
		open_out(p, (char *)0); /* This closes current	*/
	  	continue;
	  case 'h' :                   	/* Tell the SIMPLE file	*/
		if (strcmp(p, "help") == 0) { help(); exit(0); }
	  	pVal("Missing -h(eader SIMPLE) file");
	  	SIMPLE_file = p;
	  	continue;
	  case 'i' :                   	/* Tell which "ReadMe" 	*/
		pVal("Missing -i FileName");
    		tinfo = cad_new(p, fop);	/* Format File 	*/
		intro = p; i = strbloc(p, dirsep);
#if 0
		if ((!ddat) && (i>=0)) {  	/* Keep  Path 	*/
		    ddat = strdup(intro);
		    ddat[i+1] = 0;
		}
#endif
		continue;
	  case 'm' :                   	/* Maximum records...	*/
		pVal("Missing -m(ax_records) value");
		maxrecs = atol(p) ;
    		continue ;
	  case 'v':                     /* Verbose      */
		if (strncmp(p, "ver", 3) == 0) {
		    printf("tofits (CDS) Version %s\n", VERSION);
		    exit(0);
		}
		verbop = 1; fop |= FOP_v;
		fprintf(stderr, "....PATH=%s\n", getenv("PATH"));
		/* osmtrace(1); */
		continue;
	  case 'z':                     /* zip Compress */
		start_out("gzip");
		continue;
	  case 'Z':                     /* Compress     */
		start_out("compress");
		continue;
	  case 'f':			/* -fih (generate from *.fih) 	*/
	  	if (strdiff(p, "fih") == 0) { 
	  	    ofih = 1;
	  	    continue; 
	    	}
	  	goto bad_option;
	  case 'n':			/* -nodata	*/
	  	if (strdiff(p, "nodata") == 0) { toFITSd = 0; continue; }
	  	if (strdiff(p, "nodat") == 0)  { toFITSd = 2; continue; }
	  	if (strdiff(p, "noh") == 0)    { toFITSh = 0; continue; }
	  	goto bad_option;
	  case 'u':			/* -u		*/
		ou = atoi(p+1);
		continue;
	  case 'w':			/* -w		*/
	  	if (strdiff(p, "w.fit") == 0) 	   ow |= 1|WOP_fit; 
	  	else if (strdiff(p, "w"    ) == 0) ow |= 1|WOP_fit; 
	  	else if (strdiff(p, "w.fih") == 0) ow |= 1|WOP_fih; 
	  	else goto bad_option;
    		if ((ow &(WOP_fit|WOP_fih)) == (WOP_fit|WOP_fih)) 
    		    terminate("Can't have options  -w   and  -w.fih");
	  	continue ;
	  case 's': case 't':		/* Use table Structure	*/
	  	pVal("Missing -t(abular tructure) qualifier") ;
	  	setup() ;
    		if (!tinfo) terminate("No ReadMe available");
		/* Specifying -t. or -t- stops the -t choice	*/
		tparm = isalnum(*p) ? p : (char *)0 ;
		continue ;
	  case '\0':			/* - alone		*/
    	    	if (tparm) convert(tparm, stdin) ;
		nf++ ;			/* Don't process tables	*/
		continue ;
	  default:
	  bad_option:
		fprintf(stderr, "\n****Bad option: -%s\n", p);
		errors++;
		terminate((char *)0);
	} 
	
	/*----------- A file to Process ---------------------------------*/

	/* A special file asks to stop file processing */
	if (((*p == '.')&&p[1]) || (*p == '$')) {
	    nf++ ;
	    continue ;
	}

	/* V3.23: separate dir/./file */
	while ((d = strstr(p, "/./"))) { char *s;
	    s = strdup(p); s[(d-p)] = 0;
	    cd(s); free(s);
	    p = d+3;
	}
	
	if (is_dir(p)) {
	    cd(p) ;
	    continue ;
	}

	setup() ;
	/* In order to have fzopen working, the .gz or .Z suffix must
	   be removed...
	*/
	i = strlen(p) -1;
	if (tolower(p[i]) == 'z') {
	    i-- ;
	    if (p[i] == '.') p[i] = 0;
	    else if (strcmp(p+i-1, ".gz") == 0) {
		p[--i] = 0;
	    }
	}
	if (tparm) {			/* Get actual file	*/
    	    input = (*p == '-' ? stdin : fzopen(p)) ;
    	    if (!input) error(p),
    		terminate("Can't open file in -t option") ;
    	    convert(tparm, input) ;
	}
	else convert(p, (FILE *)0) ;
	nf++ ;
    }

    /*---------------------------- End of loop on Arguments ------------*/

    if (nf == 0) {		/* Convert by default all files --------*/
    
    	/* Process now every file 	*/
    	setup() ;
    	if (!tinfo) terminate("No file to process");
	if (tparm) convert(tparm, stdin) ;
    	else for (pf = tinfo->afinfo, nf = tinfo->nfinfo; --nf >= 0; pf++) {
#if DEBUG
	    printf("\n....Starting a Description with nf=%d\n", nf);
    	    for (i=0; i<ITEMS(FITSenv); i++) 
    	        fprintf(stderr, "....Env: %s=%d\n", FITSenv[i], FITSvar[i]);
	    inbyte = getchar();
#endif
	    convert(pf->dfile, (FILE *)0) ;
    	}
    }

    open_out((char *)0, (char *)0);	/* Closes opened files */
    if (errors)   terminate((char *)0);
#if DEBUG
    	printf("====Will kill structures===="); inbyte = getchar();
#endif
    cad_kill(tinfo);

    /* Don't forget to wait until children are correctly terminated */
    if (zip_pgm[0]) wait_pipe(0);
    exit(0);
}
